from pathlib import Path
from tempfile import gettempdir


def get_temp_dir() -> Path:
    """Return the default temporary directory.

    This function fixes the problem on Windows CI where `tempfile.gettempdir`
    would return a DOS short path, not a Windows long path.

    See:
        https://learn.microsoft.com/en-us/windows/win32/fileio/naming-a-file#short-vs-long-names

    Returns:
        pathlib.Path: Long path to the default temporary directory.
    """
    return Path(gettempdir()).resolve()



from pathlib import Path
from unittest import TestCase
from unittest.mock import PropertyMock, call, patch

from dakara_base.directory import PlatformDirs

from dakara_player import user_resources


class CopyResourceTestCase(TestCase):
    """Test the copy_resource function."""

    @patch("dakara_player.user_resources.copy", autospec=True)
    @patch("dakara_player.user_resources.files", autospec=True)
    @patch.object(Path, "mkdir", autospec=True)
    @patch.object(Path, "exists", autospec=True)
    def test_copy(
        self,
        mocked_exists,
        mocked_mkdir,
        mocked_files,
        mocked_copy,
    ):
        """Test to copy files in a non existing directory."""
        mocked_exists.return_value = False
        mocked_files.return_value.iterdir.return_value = [
            Path("package/resources/file1.ext"),
            Path("package/resources/file2.ext"),
            Path("package/resources/__init__.py"),
        ]

        user_resources.copy_resource("package.resources", Path("destination"), False)

        mocked_exists.assert_called_with(Path("destination"))
        mocked_mkdir.assert_called_with(
            Path("destination"), parents=True, exist_ok=True
        )
        mocked_files.assert_called_with("package.resources")
        mocked_copy.assert_has_calls(
            [
                call(Path("package/resources/file1.ext"), Path("destination")),
                call(Path("package/resources/file2.ext"), Path("destination")),
            ]
        )

    @patch("dakara_player.user_resources.input")
    @patch("dakara_player.user_resources.copy", autospec=True)
    @patch("dakara_player.user_resources.files", autospec=True)
    @patch.object(Path, "mkdir", autospec=True)
    @patch.object(Path, "exists", autospec=True)
    def test_copy_existing_abort(
        self,
        mocked_exists,
        mocked_mkdir,
        mocked_files,
        mocked_copy,
        mocked_input,
    ):
        """Test to copy files in an existing directory and abort."""
        mocked_exists.return_value = True
        mocked_files.return_value.iterdir.return_value = [
            Path("package/resources/file1.ext"),
            Path("package/resources/file2.ext"),
            Path("package/resources/__init__.py"),
        ]
        mocked_input.return_value = "no"

        user_resources.copy_resource("package.resources", Path("destination"), False)

        mocked_exists.assert_called_with(Path("destination"))
        mocked_mkdir.assert_not_called()
        mocked_files.assert_not_called()
        mocked_copy.assert_not_called()

    @patch("dakara_player.user_resources.input")
    @patch("dakara_player.user_resources.copy", autospec=True)
    @patch("dakara_player.user_resources.files", autospec=True)
    @patch.object(Path, "mkdir", autospec=True)
    @patch.object(Path, "exists", autospec=True)
    def test_copy_existing_overwrite(
        self,
        mocked_exists,
        mocked_mkdir,
        mocked_files,
        mocked_copy,
        mocked_input,
    ):
        """Test to copy files in an existing directory and overwrite."""
        mocked_exists.return_value = True
        mocked_files.return_value.iterdir.return_value = [
            Path("package/resources/file1.ext"),
            Path("package/resources/file2.ext"),
            Path("package/resources/__init__.py"),
        ]
        mocked_input.return_value = "yes"

        user_resources.copy_resource("package.resources", Path("destination"), False)

        mocked_exists.assert_called_with(Path("destination"))
        mocked_mkdir.assert_called_with(
            Path("destination"), parents=True, exist_ok=True
        )
        mocked_files.assert_called_with("package.resources")
        mocked_copy.assert_has_calls(
            [
                call(Path("package/resources/file1.ext"), Path("destination")),
                call(Path("package/resources/file2.ext"), Path("destination")),
            ]
        )

    @patch("dakara_player.user_resources.input")
    @patch("dakara_player.user_resources.copy", autospec=True)
    @patch("dakara_player.user_resources.files", autospec=True)
    @patch.object(Path, "mkdir", autospec=True)
    @patch.object(Path, "exists", autospec=True)
    def test_copy_existing_force(
        self,
        mocked_exists,
        mocked_mkdir,
        mocked_files,
        mocked_copy,
        mocked_input,
    ):
        """Test to force copy files in an existing directory."""
        mocked_exists.return_value = True
        mocked_files.return_value.iterdir.return_value = [
            Path("package/resources/file1.ext"),
            Path("package/resources/file2.ext"),
            Path("package/resources/__init__.py"),
        ]

        user_resources.copy_resource("package.resources", Path("destination"), True)

        mocked_input.assert_not_called()
        mocked_exists.assert_not_called()
        mocked_mkdir.assert_called_with(
            Path("destination"), parents=True, exist_ok=True
        )
        mocked_files.assert_called_with("package.resources")
        mocked_copy.assert_has_calls(
            [
                call(Path("package/resources/file1.ext"), Path("destination")),
                call(Path("package/resources/file2.ext"), Path("destination")),
            ]
        )


@patch.object(PlatformDirs, "user_data_path", new_callable=PropertyMock)
@patch("dakara_player.user_resources.copy_resource", autospec=True)
class CreateResourceFilesTestCase(TestCase):
    """Test the create_resource_files function."""

    @patch.object(Path, "mkdir", autospec=True)
    def test_create(self, mocked_mkdir, mocked_copy_resource, mocked_user_data_path):
        """Test to create resource files."""
        mocked_user_data_path.return_value = Path("directory")
        with self.assertLogs("dakara_player.user_resources", "DEBUG") as logger:
            user_resources.create_resource_files()

        mocked_mkdir.assert_called_with(Path("directory"), parents=True, exist_ok=True)
        mocked_copy_resource.assert_has_calls(
            [
                call(
                    "dakara_player.resources.backgrounds",
                    Path("directory") / "backgrounds",
                    False,
                ),
                call(
                    "dakara_player.resources.templates",
                    Path("directory") / "templates",
                    False,
                ),
            ]
        )

        self.assertListEqual(
            logger.output,
            [
                "INFO:dakara_player.user_resources:Resource files "
                "created in 'directory'"
            ],
        )



"""Load fonts on user level for the media players."""

import ctypes
import logging
import platform
from abc import ABC, abstractmethod
from importlib.resources import as_file, files
from pathlib import Path
from shutil import copy

from dakara_base.exceptions import DakaraError

logger = logging.getLogger(__name__)

FONT_EXTENSIONS = (".ttf", ".otf")


def get_font_loader_class():
    """Get the font loader associated to the current platform.

    Returns:
        FontLoader: Specialized version of the font loader class.
    """
    system = platform.system()

    if system == "Linux":
        return FontLoaderLinux

    if system == "Windows":
        return FontLoaderWindows

    raise NotImplementedError(
        "This operating system ({}) is not currently supported".format(system)
    )


class FontLoader(ABC):
    """Abstract font loader.

    Must be specialized for a given OS.

    Args:
        package (str): Package checked for font files.

    Attributes:
        package (str): Package checked for font files.
    """

    GREETINGS = "Abstract font loader selected"

    def __init__(self, package):
        self.package = package

        # show type of font loader
        logger.debug(self.GREETINGS)

    @abstractmethod
    def load(self):
        """Load the fonts."""

    @abstractmethod
    def unload(self):
        """Unload the loaded fonts."""

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.unload()

    def get_font_name_list(self):
        """Give name of package fonts.

        Returns:
            list of str: List of font names.
        """
        logger.debug("Scanning fonts directory")
        font_file_name_list = [
            file.name
            for file in files(self.package).iterdir()
            if file.is_file() and file.suffix.lower() in FONT_EXTENSIONS
        ]
        logger.debug("Found %i font(s) to load", len(font_file_name_list))

        return font_file_name_list

    def get_font_path_iterator(self):
        """Give font paths in font package.

        Yields:
            pathlib.Path: Absolute path to the font, from the package.
        """
        for font_file_name in self.get_font_name_list():
            with as_file(
                files(self.package).joinpath(font_file_name)
            ) as font_file_path:
                yield font_file_path


class FontLoaderLinux(FontLoader):
    """Font loader for Linux.

    It copies fonts to load in the user fonts directory and removes them on
    exit. Using symbolic links is not safe as the location of the package fonts
    may not be permanent (see `importlib.resources.path` for more info).

    See:
        https://docs.python.org/3/library/importlib.html#importlib.resources.path

    Example of use:

    >>> with FontLoaderLinux() as loader:
    ...     loader.load()
    ...     # do stuff while fonts are loaded
    >>> # now fonts are unloaded

    Args:
        package (str): Package checked for font files.

    Attributes:
        package (str): Package checked for font files.
        font_loaded (dict of pathlib.Path): List of loaded fonts. The key is
            the font file name and the value is the path of the installed font
            in user directory.
    """

    GREETINGS = "Font loader for Linux selected"
    FONT_DIR_SYSTEM = Path("/") / "usr" / "share" / "fonts"
    FONT_DIR_USER = Path("~") / ".fonts"

    def __init__(self, *args, **kwargs):
        # call parent constructor
        super().__init__(*args, **kwargs)

        # create list of fonts
        self.fonts_loaded = {}

    def get_system_font_name_list(self):
        """Retrieve the list of system fonts.

        Returns:
            list of pathlib.Path: List of font paths.
        """
        return [path.name for path in self.FONT_DIR_SYSTEM.rglob("*")]

    def get_user_font_name_list(self):
        """Retrieve the list of user fonts.

        Returns:
            list of pathlib.Path: List of font paths.
        """
        return [path.name for path in self.FONT_DIR_USER.expanduser().rglob("*")]

    def load(self):
        """Load the fonts."""
        # ensure that the user font directory exists
        self.FONT_DIR_USER.expanduser().mkdir(parents=True, exist_ok=True)

        # get system and user font files
        system_font_name_list = self.get_system_font_name_list()
        user_font_name_list = self.get_user_font_name_list()

        # load fonts
        for font_file_path in self.get_font_path_iterator():
            self.load_font(font_file_path, system_font_name_list, user_font_name_list)

    def load_font(self, font_file_path, system_font_name_list, user_font_name_list):
        """Load the provided font.

        Args:
            font_file_path (pathlib.Path): Absolute path of the font to load.
            system_font_name_list (list of pathlib.Path): List of system fonts
                name.
            user_font_name_list (list of pathlib.Path): List of user fonts
                name.
        """
        # get font file name
        font_file_name = font_file_path.name

        # check if the font is installed at system level
        if font_file_name in system_font_name_list:
            logger.debug("Font '%s' found in system directory", font_file_name)
            return

        # check if the font is installed at user level
        if font_file_name in user_font_name_list:
            logger.debug("Font '%s' found in user directory", font_file_name)
            return

        # check if the font exists as broken link at user level
        # in this case remove it and continue execution
        font_file_user_path = self.FONT_DIR_USER.expanduser() / font_file_name
        if font_file_user_path.is_symlink():
            logger.debug(
                "Dead symbolic link found for font '%s' in user directory, "
                "removing it",
                font_file_name,
            )
            font_file_user_path.unlink(missing_ok=True)

        # then, if the font is not installed, load by copying it
        copy(font_file_path, font_file_user_path)

        # register the font
        self.fonts_loaded[font_file_name] = font_file_user_path

        logger.debug(
            "Font '%s' loaded in user directory: '%s'",
            font_file_name,
            font_file_user_path,
        )

    def unload(self):
        """Remove loaded fonts."""
        for font_file_name in self.fonts_loaded.copy():
            self.unload_font(font_file_name)

    def unload_font(self, font_file_name):
        """Remove the provided font.

        Args:
            font_file_name (str): Name of the font to unload.
        """
        try:
            font_file_path = self.fonts_loaded.pop(font_file_name)
            font_file_path.unlink()
            logger.debug("Font '%s' unloaded", font_file_name)

        except OSError as error:
            logger.error(
                "Font '%s' in '%s' cannot be unloaded: %s",
                font_file_name,
                font_file_path,
                error,
            )


class FontLoaderWindows(FontLoader):
    """Font loader for Windows.

    It uses the gdi32 library to dynamically load and unload fonts for the
    current session.

    Example of use:

    >>> with FontLoaderWindows() as loader:
    ...     loader.load()
    ...     # do stuff while fonts are loaded
    >>> # fonts are unloaded

    Args:
        package (str): Package checked for font files.

    Attributes:
        package (str): Package checked for font files.
        font_loaded (dict of pathlib.Path): List of loaded fonts. The key is
            the font file name and the value is the path of font used at
            installation.
    """

    GREETINGS = "Font loader for Windows selected"

    def __init__(self, *args, **kwargs):
        # call parent constructor
        super().__init__(*args, **kwargs)

        # create list of fonts
        self.fonts_loaded = {}

        # check we can use gdi32 library
        if not hasattr(ctypes, "windll"):
            raise FontLoaderNotAvailableError(
                "FontLoaderWindows can only be used on Windows"
            )

    def load(self):
        """Load the fonts."""
        for font_file_path in self.get_font_path_iterator():
            self.load_font(font_file_path)

    def load_font(self, font_file_path):
        """Load the provided font.

        Args:
            font_file_path (pathlib.Path): Absolute path of the font to load.
        """
        success = ctypes.windll.gdi32.AddFontResourceW(font_file_path)
        if success:
            self.fonts_loaded[font_file_path.name] = font_file_path
            logger.debug("Font '%s' loaded", font_file_path.name)
            return

        logger.warning("Font '%s' cannot be loaded", font_file_path.name)

    def unload(self):
        """Remove loaded fonts."""
        for font_file_name in self.fonts_loaded.copy():
            self.unload_font(font_file_name)

    def unload_font(self, font_file_name):
        """Remove the provided font.

        Args:
            font_file_name (str): Name of the font to unload.
        """
        font_file_path = self.fonts_loaded.pop(font_file_name)
        success = ctypes.windll.gdi32.RemoveFontResourceW(font_file_path)
        if success:
            logger.debug("Font '%s' unloaded", font_file_name)
            return

        logger.warning("Font '%s' cannot be unloaded", font_file_name)


class FontLoaderNotAvailableError(DakaraError):
    """Error raised when a font loader cannot be used."""



"""Command line interface to run the player."""

import logging
import sys
from argparse import ArgumentParser

from dakara_base.config import (
    Config,
    ConfigNotFoundError,
    create_config_file,
    create_logger,
    set_loglevel,
)
from dakara_base.directory import directories
from dakara_base.exceptions import (
    DakaraError,
    generate_exception_handler,
    handle_all_exceptions,
)
from dakara_base.http_client import ParameterError

from dakara_player.player import DakaraPlayer
from dakara_player.user_resources import create_resource_files
from dakara_player.version import __date__, __version__

CONFIG_FILE = "player.yaml"
CONFIG_PREFIX = "DAKARA"


logger = logging.getLogger(__name__)
handle_config_not_found = generate_exception_handler(
    ConfigNotFoundError, "Please run 'dakara-player create-config'"
)
handle_config_incomplete = generate_exception_handler(
    DakaraError,
    "Config may be incomplete, please check '{}'".format(
        directories.user_config_path / CONFIG_FILE
    ),
)
handle_parameter_error = generate_exception_handler(
    ParameterError,
    "Config may be incomplete, please check '{}'".format(
        directories.user_config_path / CONFIG_FILE
    ),
)


def get_parser():
    """Get a parser.

    Returns:
        argparse.ArgumentParser: Parser.
    """
    # main parser
    parser = ArgumentParser(
        prog="dakara-player", description="Player for the Dakara project"
    )

    parser.set_defaults(function=lambda _: parser.print_help())

    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="enable debug output, increase verbosity",
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {} ({})".format(__version__, __date__),
    )

    # subparsers
    subparser = parser.add_subparsers(title="subcommands")

    # play subparser
    play_subparser = subparser.add_parser(
        "play",
        description="Run the player",
        help="run the player",
    )
    play_subparser.set_defaults(function=play)

    # create config subparser
    create_config_subparser = subparser.add_parser(
        "create-config",
        description="Create a new config file in user directory",
        help="create a new config file in user directory",
    )
    create_config_subparser.set_defaults(function=create_config)

    create_config_subparser.add_argument(
        "--force",
        help="overwrite previous config file if it exists",
        action="store_true",
    )

    # create resources subparser
    create_resource_subparser = subparser.add_parser(
        "create-resources",
        description="Create resource files in user directory "
        "(for background screens, text templates)",
        help="create resource files in user directory",
    )
    create_resource_subparser.set_defaults(function=create_resources)

    create_resource_subparser.add_argument(
        "--force",
        help="overwrite existing files if any",
        action="store_true",
    )

    return parser


def play(args):
    """Run the player.

    Args:
        args (argparse.Namespace): Arguments from command line.
    """
    with handle_config_not_found():
        create_logger()
        config = Config(CONFIG_PREFIX)
        config.load_file(directories.user_config_path / CONFIG_FILE)
        config.check_mandatory_keys(["player", "server"])
        config.set_debug(args.debug)
        set_loglevel(config)

    dakara = DakaraPlayer(config)

    with handle_config_incomplete():
        dakara.load()

    # catch HTTP client missing credentials
    with handle_parameter_error():
        dakara.run()


def create_config(args):
    """Create the config file.

    Args:
        args (argparse.Namespace): Arguments from command line.
    """
    create_logger(custom_log_format="%(message)s", custom_log_level="INFO")
    create_config_file("dakara_player.resources", CONFIG_FILE, args.force)
    logger.info("Please edit this file")


def create_resources(args):
    """Create resource files.

    Args:
        args (argparse.Namespace): Arguments from command line.
    """
    create_logger(custom_log_format="%(message)s", custom_log_level="INFO")
    create_resource_files(args.force)
    logger.info("You can now customize those files")


def main():
    """Main command."""
    parser = get_parser()
    args = parser.parse_args()

    with handle_all_exceptions(
        bugtracker_url="https://github.com/DakaraProject/dakara-player/issues",
        logger=logger,
        debug=args.debug,
    ) as exit_value:
        args.function(args)

    sys.exit(exit_value.value)



"""Manage a window."""

import logging
from abc import ABC, abstractmethod
from threading import Event, Thread

try:
    import tkinter

except ImportError:
    tkinter = None


logger = logging.getLogger(__name__)


class BaseWindowManager(ABC):
    """Abstract class for window manager.

    Args:
        title (str): Title of the window.
        fullscreen (bool): If `True`, the window is fullscreen.
        disabled (bool): If `True`, no window are displayed.
    """

    GREETINGS = "Abstract window manager selected"

    @staticmethod
    @abstractmethod
    def is_available():
        """Tell if this window manager can be initialized.

        Returns:
            bool: `True` if the window managen can be initialized.


        Raises:
            WindowManagerNotAvailableError: If no window manager is available.
        """

    def __init__(self, title="No title", fullscreen=False, disabled=False):
        if not self.is_available():
            raise WindowManagerNotAvailableError(
                "This window manager cannot be initialized"
            )

        logger.debug(self.GREETINGS)
        self.title = title
        self.fullscreen = fullscreen
        self.disabled = disabled

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    @abstractmethod
    def open(self):
        """Open the window."""

    @abstractmethod
    def close(self):
        """Close the window."""

    @abstractmethod
    def get_id(self):
        """Get window ID.

        Retuns:
            int: ID of the window.
        """


class DummyWindowManager(BaseWindowManager):
    """Dummy window manager.

    It never creates a window and is always available.

    Args:
        title (str): Title of the window.
        fullscreen (bool): If `True`, the window is fullscreen.
        disabled (bool): If `True`, no window are displayed.
    """

    GREETINGS = "Dummy window manager selected"

    @staticmethod
    def is_available():
        """Tell if this window manager can be initialized.

        Returns:
            bool: `True` if the window managen can be initialized.
        """
        return True

    def open(self):
        """Open the window."""

    def close(self):
        """Close the window."""

    def get_id(self):
        """Get window ID.

        Retuns:
            None: The ID is always `None`.
        """
        return None


class TkWindowManager(BaseWindowManager):
    """Tkinter window manager.

    Uses the Python default GUI library Tkinter.

    The window is created in a separate thread and checks every
    `ACTUALIZE_INTERVAL` if it has to close (using an event set by the `close`
    method). This indirect approach works well, even if the window doesn't
    close immediately at shutdown. A more direct approach was tried using
    `generate_event`, but this messes up the destruction order and leads to a
    crash of the interpreter. Using `quit` instead of `destroy` as a workaround
    was problematic for tests, as the method doesn't cleanup after itself,
    making impossible to close a second window.

    See:
        https://stackoverflow.com/q/66529633/4584444

    Args:
        title (str): Title of the window.
        fullscreen (bool): If `True`, the window is fullscreen.
        disabled (bool): If `True`, no window are displayed.
    """

    GREETINGS = "Tk window selected"
    ACTUALIZE_INTERVAL = 1000

    @staticmethod
    def is_available():
        """Tell if this window manager can be initialized.

        Returns:
            bool: `True` if the window managen can be initialized.
        """
        return tkinter is not None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.window_thread = Thread(target=self.create_window)
        self.stop = Event()
        self.ready = Event()
        self.id = None

    def create_window(self):
        """Create the window.

        Must be called in its own thread.
        """

        class TkWindow(tkinter.Tk):
            """Custom Tk window."""

            def __init__(self, stop, title, fullscreen, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.stop = stop
                self.title(title)
                self.configure(width=1024, height=576, bg="black", cursor="none")
                self.resizable(True, True)
                self.attributes("-fullscreen", fullscreen)
                self.after(TkWindowManager.ACTUALIZE_INTERVAL, self.actualize)

            def actualize(self):
                """Check if the window has to be closed periodically."""
                if self.stop.is_set():
                    self.destroy()
                    return

                self.after(TkWindowManager.ACTUALIZE_INTERVAL, self.actualize)

        logger.debug("Creating Tk window thread")
        window = TkWindow(self.stop, self.title, self.fullscreen)
        self.id = window.winfo_id()

        # allow to not draw the window for testing purposes
        if self.disabled:
            window.withdraw()

        self.ready.set()
        window.mainloop()

    def open(self):
        """Open the window."""
        logger.debug("Creating Tk window")
        self.window_thread.start()

    def close(self):
        """Close the window."""
        self.ready.wait()
        logger.debug("Closing Tk window")
        self.stop.set()
        self.window_thread.join()

    def get_id(self):
        """Get window ID.

        Retuns:
            int: ID of the window.
        """
        self.ready.wait()
        logger.debug("Getting Tk window ID")
        return self.id


def get_window_manager_class():
    """Give an available window manager class.

    Returns:
        BaseWindowManager: Tries to return `TkWindowManager`, fallbacks to
        `DummyWindowManager`.
    """
    if TkWindowManager.is_available():
        return TkWindowManager

    return DummyWindowManager


WindowManager = get_window_manager_class()
"""BaseWindowManager: Available window manager class.
"""


class WindowManagerNotAvailableError(Exception):
    """Error when initializing an unavailable window manager."""



"""Version of the program.

The version and date module variables are automatically updated by
`bump_version.sh`.
"""

import logging

from packaging.version import parse

__version__ = "1.9.0"
__date__ = "2025-03-06"

logger = logging.getLogger(__name__)


def check_version():
    """Display version number and check if on release."""
    # log player versio
    logger.info("Dakara player %s (%s)", __version__, __date__)

    # check version is a release
    version = parse(__version__)
    if version.is_prerelease:
        logger.warning("You are running a dev version, use it at your own risks!")



"""mpv media player."""

import logging
import re
from abc import ABC
from pathlib import Path

from dakara_base.exceptions import DakaraError
from dakara_base.safe_workers import safe
from packaging.version import Version, parse

try:
    import python_mpv_jsonipc as mpv

except ImportError:
    mpv = None

from dakara_player.media_player.base import (
    InvalidStateError,
    MediaPlayer,
    VersionNotFoundError,
    on_playing_this,
)

logger = logging.getLogger(__name__)
mpv_logger = logging.getLogger("mpv")

SUBTITLE_EXTENSIONS = [
    ".ass",
    ".ssa",
]

MPV_ERROR_LEVELS = {
    "fatal": logging.CRITICAL,
    "error": logging.ERROR,
    "warn": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}

PLAYER_IS_AVAILABLE_ATTEMPTS = 5

USE_PATH_AUDIO = -1


# monkey patch mpv to silent socket close failures on windows
if mpv is not None:

    class WindowsSocketSilenced(mpv.WindowsSocket):
        def stop(self, *args, **kwargs):
            try:
                super().stop(*args, **kwargs)

            except OSError:
                pass

    mpv.WindowsSocket = WindowsSocketSilenced


class MediaPlayerMpv(MediaPlayer, ABC):
    """Abstract class to manipulate mpv.

    This class contains the minimum static methods to detect mpv availability
    and installed version. It can be used to instanciate the correct version
    with the static method `from_version`.
    """

    player_name = "mpv"

    @staticmethod
    def is_available():
        """Indicate if mpv is available.

        Try the detection `PLAYER_IS_AVAILABLE_ATTEMPTS` times.

        Returns:
            bool: `True` if mpv is useable.
        """
        if mpv is None:
            return False

        for _ in range(PLAYER_IS_AVAILABLE_ATTEMPTS):
            try:
                player = mpv.MPV()
                player.terminate()
                return True

            except FileNotFoundError:
                pass

        return False

    @staticmethod
    def get_version():
        """Get media player version.

        mpv versions are in the form "vMAJOR.MINOR.PATCH" for releases, older
        versions (<0.37) do not include a 'v' before the version, post releases
        are detected as any valid base version followed by "+BUILD" or "-BUILD"
        where BUILD is whatever build information was included in the string.

        Returns:
            packaging.version.Version: Parsed version of mpv.

        Raises:
            VersionNotFoundError: If unable to parse version.
        """
        player = mpv.MPV()
        match = re.search(
            r"mpv v?(\d+\.\d+\.\d+)([+-]\w+)?",
            player.mpv_version,
        )
        player.terminate()

        if match:
            if match.group(2):
                return parse(match.group(1) + "-post")

            return parse(match.group(1))

        raise VersionNotFoundError("Unable to get mpv version")

    @staticmethod
    def get_class_from_version(version):
        """Get the mpv media player class according to installed version.

        Args:
            version (packaging.version.Version): Arbitrary mpv version to use.

        Returns:
            MediaPlayerMpv: Will return the class adapted to the version of mpv:

                - `MediaPlayerMpvPost0340` if mpv newer than 0.34.0;
                - `MediaPlayerMpvPost0330` if mpv newer than 0.33.0;
                - `MediaPlayerMpvOld` as default.

        Raises:
            MpvTooOldError: if MPV version is lower than 0.28.0
        """
        if version >= Version("0.34.0"):
            logger.debug("Using post 0.34.0 API of mpv")
            return MediaPlayerMpvPost0340

        if version >= Version("0.33.0"):
            logger.debug("Using post 0.33.0 API of mpv")
            return MediaPlayerMpvPost0330

        if version < Version("0.28.0"):
            raise MpvTooOldError(
                f"MPV is too old ({version=}, version 0.28.0 and higher supported)"
            )

        logger.debug("Using old API of mpv")
        return MediaPlayerMpvOld

    @staticmethod
    def from_version(*args, **kwargs):
        """Instanciate the right mpv media player class.

        Returns:
            MediaPlayer: Instance of the mpv media player for the correct
            version of mpv.
        """
        try:
            config = kwargs.get("config") or args[2]
            version = Version(config["mpv"]["force_version"])

        except (KeyError, IndexError):
            version = MediaPlayerMpv.get_version()

        return MediaPlayerMpv.get_class_from_version(version)(*args, **kwargs)


class MediaPlayerMpvOld(MediaPlayerMpv):
    """Class to manipulate old mpv versions (< 0.33.0).

    The class can be used as a context manager that closes mpv
    automatically on exit.

    Any exception in callbacks make the application to crash.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        config (dict): Dictionary of configuration.
        tempdir (pathlib.Path): Path of the temporary directory.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        player_name (str): Name of mpv.
        fullscreen (bool): If `True`, mpv will be fullscreen.
        kara_folder_path (pathlib.Path): Path to the karaoke folder.
        playlist_entry (dict): Playlist entyr object.
        callbacks (dict): High level callbacks associated with the media
            player.
        warn_long_exit (bool): If `True`, display a warning message if the media
            player takes too long to stop.
        durations (dict of int): Duration of the different screens in seconds.
        text_paths (dict of pathlib.Path): Path of the different text screens.
        text_generator (dakara_player.text_generator.TextGenerator): Text
            generator instance.
        background_loader
        (dakara_player.background_loader.BackgroundLoader): Background
            loader instance.
        player (mpv.MPV): Instance of mpv.
        playlist_entry_data (dict): Extra data of the playlist entry.
        player_data (dict): Extra data of the player.
    """

    def init_player(self, config, tempdir):
        """Initialize the objects of mpv.

        Actions performed in this method should not have any side effects
        (query file system, etc.).

        Args:
            config (dict): Dictionary of configuration.
            tempdir (pathlib.Path): Path of the temporary directory.
        """
        # set mpv player options and logging
        loglevel = config.get("loglevel", "info")
        self.player = mpv.MPV(log_handler=self.handle_log_messages, loglevel=loglevel)
        config_mpv = config.get("mpv") or {}

        for key, value in config_mpv.items():
            try:
                self.player.__setattr__(key, value)

            except mpv.MPVError:
                logger.error(f"Unable to set mpv option '{key}' to value '{value}'")

        # playlist entry objects
        self.playlist_entry_data = {}
        self.clear_playlist_entry_player()

        # player objects
        self.player_data = {}
        self.player_data["skip"] = False

    def load_player(self):
        """Perform actions with side effects for mpv initialization."""
        # set mpv callbacks
        self.set_mpv_default_callbacks()

        # set mpv fullscreen
        self.player.fullscreen = self.fullscreen

        # log mpv version
        logger.info("mpv %s", self.get_version_str())

        # set mpv as a single non-interactive window
        self.player.force_window = "immediate"
        self.player.input_default_bindings = False
        self.player.osc = False
        self.player.osd_level = 0

        # set window title
        self.player.title = "Dakara player mpv"

        # handle image transitions as videos
        self.player.demuxer_lavf_o = "loop=1"
        # used for idle/transitions screens
        self.player.image_display_duration = "inf"

    @on_playing_this(["song"], default_return=0)
    def get_timing(self):
        """Get mpv timing.

        Returns:
            int: Current song timing in seconds if a song is playing, or 0 when
                idle or during transition screen.
        """
        timing = self.player.time_pos or 0

        return int(timing)

    def set_mpv_default_callbacks(self):
        """Set mpv default callbacks."""
        self.player.bind_event("end-file", self.handle_end_file)
        self.player.bind_event("start-file", self.handle_start_file)
        self.player.bind_event("pause", self.handle_pause)
        self.player.bind_event("unpause", self.handle_unpause)

    def is_playing(self):
        """Query if mpv is playing something.

        Returns:
            bool: `True` if mpv is playing something.

        Raises:
            AssertError: If too many intries are present in the mpv internal
                playlist.
        """
        # query if the player is currently playing
        if self.is_paused():
            return False

        playlist = self.player.playlist

        if len(playlist) == 0:
            return False

        assert len(playlist) == 1, "Too many entries in mpv internal playlist"

        media = playlist[0]
        return media.get("playing", False)

    def is_paused(self):
        """Query if mpv is paused.

        Returns:
            bool: `True` if mpv is paused.
        """
        return self.player.pause

    def is_playing_this(self, what):
        """Query if mpv is playing the requested media type.

        It is pretty difficult to get what mpv is/was playing, as it does not
        have a media object, but only a path to the current media file (that
        disappears when the said file ends), and a playlist that contains the
        path of the latest media file (that disappears when another file is set
        to play). We can only rely on the playlist path, and this is pretty
        weak. I don't have any better solution for now.

        Args:
            what (str): Tell if mpv current track is of the requested type, but
                not if it is actually playing it (it can be in pause).

        Returns:
            bool: `True` if mpv is playing the requested type.

        Raises:
            AssertError: If too many entries are present in the mpv internal
                playlist.
        """
        playlist = self.player.playlist

        if len(playlist) == 0:
            return False

        assert len(playlist) == 1, "Too many entries in mpv internal playlist"

        media_raw = playlist[0].get("filename")

        if media_raw is None:
            return False

        media_path = Path(media_raw)

        if what == "idle":
            return media_path == self.background_loader.backgrounds["idle"]

        return media_path == self.playlist_entry_data[what].path

    def play(self, what):
        """Request mpv to play something.

        No preparation should be done by this function, i.e. the media track
        should have been prepared already by `set_playlist_entry`.

        Args:
            what (str): What media to play.

        Raises:
            ValueError: If the action to play is unknown.
        """
        # reset player
        self.player.sub_files = []
        self.player.audio_files = []
        self.player.audio = "auto"
        self.player.pause = False
        self.player.end = "none"

        if what == "idle":
            # if already idle, do nothing
            if self.is_playing_this("idle"):
                return

            self.generate_text("idle")
            self.player.play(str(self.background_loader.backgrounds["idle"]))
            self.player.sub_files = str(self.text_paths["idle"])

            return

        if what == "transition":
            self.player.play(str(self.playlist_entry_data["transition"].path))
            self.player.sub_files = str(self.text_paths["transition"])
            self.player.end = str(self.durations["transition"])

            return

        if what == "song":
            # manage instrumental track/file
            track_id_audio = self.playlist_entry_data["song"].track_id_audio
            if track_id_audio is not None:
                if track_id_audio == USE_PATH_AUDIO:
                    path_audio = self.playlist_entry_data["song"].path_audio
                    self.player.audio_files = [str(path_audio)]
                    logger.debug("Requesting to play audio file %s", path_audio)

                else:
                    self.player.audio = track_id_audio
                    logger.debug("Requesting to play audio track %i", track_id_audio)

            # if the subtitle file cannot be discovered, do not request it
            if self.playlist_entry_data["song"].path_subtitle:
                self.player.sub_files = [
                    str(self.playlist_entry_data["song"].path_subtitle)
                ]

            self.player.play(str(self.playlist_entry_data["song"].path))

            return

        raise ValueError("Unexpected action to play: {}".format(what))

    @on_playing_this(["transition", "song"])
    def pause(self):
        """Request mpv to pause or unpause.

        Can only work on transition screens or songs. Pausing should have no
        effect if mpv is already paused.
        """
        if self.is_paused():
            logger.debug("Player already in pause")
            return

        logger.info("Setting pause")
        self.player.pause = True

    @on_playing_this(["transition", "song"])
    def resume(self):
        """Request mpv to resume playing.

        Can only work on transition screens or songs. Resuming should have no
        effect if mpv is already playing.
        """
        if not self.is_paused():
            logger.debug("Player already playing")
            return

        logger.info("Resuming play")
        self.player.pause = False

    @on_playing_this(["song"])
    def restart(self):
        """Request to restart the current media.

        Can only work on songs.
        """
        logger.info("Restarting media")
        self.player.time_pos = 0
        self.callbacks["updated_timing"](self.playlist_entry["id"], self.get_timing())

    @on_playing_this(["transition", "song"])
    def skip(self, no_callback=False):
        """Request to skip the current media.

        Can only work on transition screens or songs. mpv should
        continue playing, but media has to be considered already finished.

        Args:
            no_callback (bool): If `True`, no callback to signal the song has
                finished will be executed.
        """
        logger.info("Skipping '%s'", self.playlist_entry["song"]["title"])
        self.player_data["skip"] = True
        if not no_callback:
            self.callbacks["finished"](self.playlist_entry["id"])

        self.clear_playlist_entry()

    @on_playing_this(["song"])
    def rewind(self):
        """Request to rewind a few seconds the media.

        Can only work on songs. It cannot rewind before the beginning of the
        media. In that case, restart the song.
        """
        timing = self.player.time_pos - self.durations["rewind_fast_forward"]

        if timing < 0:
            self.restart()
            return

        logger.info("Rewinding media")
        self.player.time_pos = timing
        self.callbacks["updated_timing"](self.playlist_entry["id"], self.get_timing())

    @on_playing_this(["song"])
    def fast_forward(self):
        """Request to fast forward a few seconds the media.

        Can only work on songs. It cannot advance passed the end of the media.
        In that case, skip the song.
        """
        timing = self.player.time_pos + self.durations["rewind_fast_forward"]

        if timing > self.player.duration:
            self.skip()
            return

        logger.info("Fast forwarding media")
        self.player.time_pos = timing
        self.callbacks["updated_timing"](self.playlist_entry["id"], self.get_timing())

    def stop_player(self):
        """Request to stop mpv."""
        logger.info("Stopping player")
        self.player.terminate()
        logger.debug("Stopped player")

    def set_playlist_entry_player(self, playlist_entry, file_path, autoplay):
        """Prepare playlist entry data to be played.

        Prepare all media objects, subtitles, etc. for being played, for the
        transition screen and the song. Such data are stored on the dedicated
        object `playlist_entry_data`.

        Args:
            playlist_entry (dict): Playlist entry object.
            file_path (pathlib.Path): Absolute path to the song file.
            autoplay (bool): If `True`, start to play transition screen as soon
                as possible (i.e. as soon as the transition screen media is
                ready). The song media is prepared when the transition screen
                is playing.
        """
        # if the player is playing the idle screen, mark to skip it
        if self.is_playing_this("idle"):
            self.player_data["skip"] = True

        # set transition
        self.playlist_entry_data["transition"].path = (
            self.background_loader.backgrounds["transition"]
        )
        self.generate_text("transition")

        if autoplay:
            self.play("transition")

        # set song
        self.playlist_entry_data["song"].path = file_path

        # manually set the subtitles as a workaround for the matching of
        # mpv being too permissive
        path_without_ext = file_path.parent / file_path.stem
        for subtitle_extension in SUBTITLE_EXTENSIONS:
            path_subtitle = path_without_ext.with_suffix(subtitle_extension)
            if path_subtitle.exists():
                break

        else:
            path_subtitle = None

        self.playlist_entry_data["song"].path_subtitle = path_subtitle

        # manage instrumental
        if playlist_entry["use_instrumental"]:
            self.manage_instrumental(playlist_entry, file_path)

    def manage_instrumental(self, playlist_entry, file_path):
        """Manage the requested instrumental track.

        Instrumental track is searched first in audio files having the same
        name as the video file, then in extra audio tracks of the video file.

        As mpv cannot fetch information of a media in advance, we have to
        discover and set the instrumental track when the media starts.

        Args:
            playlist_entry (dict): Playlist entry data. Must contain the key
                `use_instrumental`.
            file_path (pathlib.Path): Path of the song file.
        """
        # get instrumental file if possible
        audio_path = self.get_instrumental_file(file_path)

        if audio_path:
            self.playlist_entry_data["song"].track_id_audio = USE_PATH_AUDIO
            self.playlist_entry_data["song"].path_audio = audio_path
            logger.info(
                "Requesting to play instrumental file '%s' for '%s'",
                audio_path,
                file_path,
            )

            return

        # otherwise mark to use the 2nd track when starting to read the media
        # mpv use different index for each track, so we can safely request the
        # second audio track
        self.playlist_entry_data["song"].track_id_audio = 2
        logger.info("Requesting to play instrumental track of '%s'", file_path)

    def clear_playlist_entry_player(self):
        """Clean playlist entry data after being played."""
        self.playlist_entry_data = {
            "transition": Media(),
            "song": MediaSong(),
        }

    @safe
    def handle_end_file(self, event):
        """Callback called when a media ends.

        This happens when:
            - A transition screen ends, leading to playing the actual song;
            - A song ends normally, leading to calling the callback
                `callbacks["finished"]`;
            - A song ends because it has been skipped, this case is ignored.

        Args:
            event (dict): Mpv event.

        Raises:
            InvalidStateError: If the context of call of this callback is
                unexpected.
        """
        logger.debug("File end callback called")

        # only handle when a file naturally ends (i.e. is not skipped)
        # i know this strategy is risky but it is the only way for this version
        # of mpv
        if self.player_data["skip"]:
            self.player_data["skip"] = False
            logger.debug("File has been skipped")

            return

        # the transition screen has finished, request to play the song itself
        if self.is_playing_this("transition"):
            logger.debug("Will play '{}'".format(self.playlist_entry_data["song"].path))
            self.play("song")

            return

        # the media has finished, so clean memory and call the according callback
        if self.is_playing_this("song"):
            playlist_entry_id = self.playlist_entry["id"]
            self.clear_playlist_entry()
            self.callbacks["finished"](playlist_entry_id)

            return

        # if no state can be determined, raise an error
        raise InvalidStateError("End file on an undeterminated state")

    @safe
    def handle_log_messages(self, loglevel, component, message):
        """Callback called when a log message occurs.

        Direct the message to the logger for Dakara Player. If the level is
        "fatal" or higher, call the callback `callbacks["error"]` and skip the
        media.

        Args:
            loglevel (str): Level of the log message.
            component (str): Component of mpv that generated the message.
            message (str): Actual log message.
        """
        logger.debug("Log message callback called")
        intlevel = MPV_ERROR_LEVELS.get(loglevel, logging.NOTSET)

        # use a proper logger for mpv logs
        mpv_logger.log(intlevel, "%s: %s", component, message)

        # handle all errors here
        if intlevel == logging.CRITICAL:
            if self.is_playing_this("song"):
                logger.error("Unable to play '%s'", self.player.path)
                self.callbacks["error"](
                    self.playlist_entry["id"],
                    "Unable to play current song: {}".format(message),
                )
                self.skip()

    @safe
    def handle_start_file(self, event):
        """Callback called when a media starts.

        This happens when:
            - A transition screen starts;
            - A song starts;
            - A idle screen starts.

        Args:
            event (dict): Mpv event.

        Raises:
            InvalidStateError: If the context of call of this callback is
                unexpected.
        """
        logger.debug("Start file callback called")

        # the transition screen starts to play
        if self.is_playing_this("transition"):
            self.callbacks["started_transition"](self.playlist_entry["id"])
            logger.info(
                "Playing transition for '%s'", self.playlist_entry["song"]["title"]
            )

            return

        # the song starts to play
        if self.is_playing_this("song"):
            self.callbacks["started_song"](self.playlist_entry["id"])
            logger.info(
                "Now playing '%s' ('%s')",
                self.playlist_entry["song"]["title"],
                self.player.path,
            )

            return

        # the idle screen starts to play
        if self.is_playing_this("idle"):
            logger.debug("Playing idle screen")

            return

        raise InvalidStateError("Start file on an undeterminated state")

    def get_track_id_audio_list(self):
        """Get ID of audio tracks for the current media.

        Returns:
            list of int: ID of audio tracks in the media.
        """
        audio = [
            item["id"] for item in self.player.track_list if item["type"] == "audio"
        ]

        return audio

    @safe
    def handle_pause(self, event):
        """Callback called when paused.

        Args:
            event (dict): Mpv event.
        """
        logger.debug("Pause callback called")

        # call paused callback
        self.callbacks["paused"](self.playlist_entry["id"], self.get_timing())

        logger.debug("Paused")

    @safe
    def handle_unpause(self, event):
        """Callback called when unpaused.

        If the player is skipping a song, do not handle this event.

        Args:
            event (dict): Mpv event.
        """
        logger.debug("Unpause callback called")

        if self.player_data["skip"]:
            return

        self.callbacks["resumed"](self.playlist_entry["id"], self.get_timing())

        logger.debug("Resumed play")


class MediaPlayerMpvPost0330(MediaPlayerMpvOld):
    """Class to manipulate newer mpv versions (>= 0.33.0).

    The class can be used as a context manager that closes mpv
    automatically on exit.

    Any exception in callbacks make the application to crash.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        config (dict): Dictionary of configuration.
        tempdir (pathlib.Path): Path of the temporary directory.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        player_name (str): Name of mpv.
        fullscreen (bool): If `True`, mpv will be fullscreen.
        kara_folder_path (pathlib.Path): Path to the karaoke folder.
        playlist_entry (dict): Playlist entyr object.
        callbacks (dict): High level callbacks associated with the media
            player.
        warn_long_exit (bool): If `True`, display a warning message if the media
            player takes too long to stop.
        durations (dict of int): Duration of the different screens in seconds.
        text_paths (dict of pathlib.Path): Path of the different text screens.
        text_generator (dakara_player.text_generator.TextGenerator): Text
            generator instance.
        background_loader
        (dakara_player.background_loader.BackgroundLoader): Background
            loader instance.
        player (mpv.MPV): Instance of mpv.
        playlist_entry_data (dict): Extra data of the playlist entry.
        player_data (dict): Extra data of the player. This attribute is not
            used by the post 0.33.0 methods.
    """

    def is_playing(self):
        """Query if mpv is playing something.

        Returns:
            bool: `True` if mpv is playing something.
        """
        # query if the player is currently playing
        if self.is_paused():
            return False

        current_entries = [e for e in self.player.playlist if e.get("playing")]
        return bool(len(current_entries))

    def was_playing_this(self, what, id):
        """Query if mpv was playing the requested internal playlist ID.

        Args:
            what (str): Type of the track.
            id (int): ID of the track in mpv internal playlist.

        Returns:
            bool: `True` if mpv was playing the requested track.

        Raises:
            AssertError: If zero or more than one entries correspond to the
                provided ID.
        """
        # extract entry from playlist
        entries = [e for e in self.player.playlist if e["id"] == id]

        assert len(entries) < 2, "There are more than one media that was playing"
        assert len(entries) > 0, "No media was playing"

        return self.is_playing_this(what, entries[0]["filename"])

    def is_playing_this(self, what, current_media_path=None):
        """Query if mpv is playing the requested media type.

        Args:
            what (str): Tell if mpv current track is of the requested type, but
                not if it is actually playing it (it can be in pause).
            current_media_path (pathlib.Path): Optional current media path.

        Returns:
            bool: `True` if mpv is playing the requested type.
        """
        media_raw = current_media_path or self.player.path

        if media_raw is None:
            return False

        media_path = Path(media_raw)

        if what == "idle":
            return media_path == self.background_loader.backgrounds["idle"]

        return media_path == self.playlist_entry_data[what].path

    @safe
    def handle_end_file(self, event):
        """Callback called when a media ends.

        This happens when:
            - A transition screen ends, leading to playing the actual song;
            - A song ends normally, leading to calling the callback
                `callbacks["finished"]`;
            - A song ends because it has been skipped, this case is ignored.

        Args:
            event (dict): Mpv event.

        Raises:
            InvalidStateError: If the context of call of this callback is
                unexpected.
        """
        logger.debug("File end callback called")
        id = event["playlist_entry_id"]

        # only handle when a file naturally ends
        # we additionnaly check the skip flag as sometimes the reason is not correct
        if event["reason"] != "eof" or self.player_data["skip"]:
            self.player_data["skip"] = False
            logger.debug("File has been skipped")

            return

        # the transition screen has finished, request to play the song itself
        if self.was_playing_this("transition", id):
            logger.debug("Will play '{}'".format(self.playlist_entry_data["song"].path))
            self.play("song")

            return

        # the media has finished, so clean memory and call the according callback
        if self.was_playing_this("song", id):
            playlist_entry_id = self.playlist_entry["id"]
            self.clear_playlist_entry()
            self.callbacks["finished"](playlist_entry_id)

            return

        # if no state can be determined, raise an error
        raise InvalidStateError("End file on an undeterminated state")


class MediaPlayerMpvPost0340(MediaPlayerMpvPost0330):
    """Class to manipulate newer mpv versions (>= 0.34.0).

    The class can be used as a context manager that closes mpv
    automatically on exit.

    Any exception in callbacks make the application to crash.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        config (dict): Dictionary of configuration.
        tempdir (pathlib.Path): Path of the temporary directory.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        player_name (str): Name of mpv.
        fullscreen (bool): If `True`, mpv will be fullscreen.
        kara_folder_path (pathlib.Path): Path to the karaoke folder.
        playlist_entry (dict): Playlist entyr object.
        callbacks (dict): High level callbacks associated with the media
            player.
        warn_long_exit (bool): If `True`, display a warning message if the media
            player takes too long to stop.
        durations (dict of int): Duration of the different screens in seconds.
        text_paths (dict of pathlib.Path): Path of the different text screens.
        text_generator (dakara_player.text_generator.TextGenerator): Text
            generator instance.
        background_loader
        (dakara_player.background_loader.BackgroundLoader): Background
            loader instance.
        player (mpv.MPV): Instance of mpv.
        playlist_entry_data (dict): Extra data of the playlist entry.
        player_data (dict): Extra data of the player. Used to store if the
            player is initializing.
    """

    def set_mpv_default_callbacks(self):
        """Set mpv default callbacks.

        Some callbacks are binded to a property change, but mpv will call them
        immediately. To prevent inappropriate callbacks to be called, we have
        to register an initializing state in memory, that will be unregistered
        as soon as possible (just after calling `load`).
        """
        # binding properties
        self.player_data["initializing"] = True
        self.player.bind_property_observer("pause", self.handle_pause)

        # binding events
        self.player.bind_event("end-file", self.handle_end_file)
        self.player.bind_event("start-file", self.handle_start_file)

    def is_initializing(self):
        """Tell if mpv is initializing.

        This function is used to prevent execution of property observer
        callbacks just after their binding.

        Returns:
            bool: `True` if initializing.
        """
        return self.player_data.get("initializing", False)

    def load_player(self):
        super().load_player()

        # ensure initialization is done
        self.player_data["initializing"] = False

    @safe
    def handle_pause(self, name, paused):
        """Callback called when paused or unpaused.

        Args:
            name (str): Name of the property that changed. Should be `"pause"`.
            paused (bool): `True` if paused, `False` otherwise.
        """
        assert name == "pause"

        logger.debug("Pause callback called")

        # invalidate call if initializing
        if self.is_initializing():
            logger.debug("Pause callback aborted")

            return

        if paused:
            # call paused callback
            self.callbacks["paused"](self.playlist_entry["id"], self.get_timing())

            logger.debug("Paused")

            return

        if self.player_data["skip"]:
            return

        self.callbacks["resumed"](self.playlist_entry["id"], self.get_timing())

        logger.debug("Resumed play")


class Media:
    """Media class."""

    def __init__(self, path=None):
        self.path = path


class MediaSong(Media):
    """Song class."""

    def __init__(
        self, *args, path_subtitle=None, path_audio=None, track_id_audio=None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.path_subtitle = path_subtitle
        self.path_audio = path_audio
        self.track_id_audio = track_id_audio


class MpvTooOldError(DakaraError):
    """Error raised if MPV is too old."""



from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from dakara_player.background import BackgroundLoader, BackgroundNotFoundError


class BackgroundLoaderTestCase(TestCase):
    """Test the loader for backgrounds."""

    @patch.object(BackgroundLoader, "get_background_path", autospec=True)
    def test_load(self, mocked_get_background_path):
        """Test to load backgrounds."""
        mocked_get_background_path.return_value = Path("destination") / "idle.png"

        # create the instance
        loader = BackgroundLoader(
            destination=Path("destination"),
            package="package",
            directory=Path("directory"),
            filenames={"idle": "idle.png"},
        )

        # pre assert that there are no backgrounds
        self.assertDictEqual(loader.backgrounds, {})

        # load the backgrounds
        with self.assertLogs("dakara_player.background", "DEBUG") as logger:
            loader.load()

        # assert the backgrounds
        self.assertDictEqual(
            loader.backgrounds, {"idle": Path("destination") / "idle.png"}
        )

        # assert logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.background:Loading backgrounds",
            ],
        )

        # assert mocks
        mocked_get_background_path.assert_called_with(loader, "idle", "idle.png")

    @patch.object(BackgroundLoader, "copy_default_background", autospec=True)
    @patch.object(BackgroundLoader, "copy_custom_background", autospec=True)
    def test_get_background_path_custom(
        self, mocked_copy_custom_background, mocked_copy_default_background
    ):
        """Test to get a custom background."""
        mocked_copy_custom_background.return_value = Path("destination") / "idle.png"

        # create the instance
        loader = BackgroundLoader(
            destination=Path("destination"),
            package="package",
            directory=Path("directory"),
        )

        # get the backgrounds
        path = loader.get_background_path("idle", "idle.png")

        # assert the backgrounds
        self.assertEqual(path, Path("destination") / "idle.png")

        # assert the call of the mocked method
        mocked_copy_custom_background.assert_called_with(loader, "idle", "idle.png")
        mocked_copy_default_background.assert_not_called()

    @patch.object(BackgroundLoader, "copy_default_background", autospec=True)
    @patch.object(BackgroundLoader, "copy_custom_background", autospec=True)
    def test_get_background_path_default(
        self, mocked_copy_custom_background, mocked_copy_default_background
    ):
        """Test to get a default background."""
        mocked_copy_default_background.return_value = Path("destination") / "idle.png"

        # create the instance
        loader = BackgroundLoader(destination=Path("destination"), package="package")

        # get the backgrounds
        path = loader.get_background_path("idle", "idle.png")

        # assert the backgrounds
        self.assertEqual(path, Path("destination") / "idle.png")

        # assert the call of the mocked method
        mocked_copy_custom_background.assert_not_called()
        mocked_copy_default_background.assert_called_with(loader, "idle", "idle.png")

    @patch.object(BackgroundLoader, "copy_default_background", autospec=True)
    @patch.object(BackgroundLoader, "copy_custom_background", autospec=True)
    def test_get_background_path_custom_not_found(
        self, mocked_copy_custom_background, mocked_copy_default_background
    ):
        """Test to get a default background after trying to get a custom one."""
        mocked_copy_custom_background.side_effect = FileNotFoundError()
        mocked_copy_default_background.return_value = Path("destination") / "idle.png"

        # create the instance
        loader = BackgroundLoader(
            destination=Path("destination"),
            package="package",
            directory=Path("directory"),
        )

        # get the backgrounds
        path = loader.get_background_path("idle", "idle.png")

        # assert the backgrounds
        self.assertEqual(path, Path("destination") / "idle.png")

        # assert the call of the mocked method
        mocked_copy_custom_background.assert_called_with(loader, "idle", "idle.png")
        mocked_copy_default_background.assert_called_with(loader, "idle", "idle.png")

    @patch.object(BackgroundLoader, "copy_default_background", autospec=True)
    @patch.object(BackgroundLoader, "copy_custom_background", autospec=True)
    def test_get_background_path_default_not_found(
        self, mocked_copy_custom_background, mocked_copy_default_background
    ):
        """Test to get a unexisting background."""
        mocked_copy_default_background.side_effect = FileNotFoundError()

        # create the instance
        loader = BackgroundLoader(
            destination=Path("destination"),
            package="package",
        )

        # get the backgrounds
        with self.assertRaisesRegex(
            BackgroundNotFoundError, "No idle background file found for 'idle.png'"
        ):
            loader.get_background_path("idle", "idle.png")

        # assert the call of the mocked method
        mocked_copy_custom_background.assert_not_called()
        mocked_copy_default_background.assert_called_with(loader, "idle", "idle.png")

    @patch("dakara_player.background.copy", autospec=True)
    def test_copy_custom_background(self, mocked_copy):
        """Test to copy a custom background."""
        mocked_copy.return_value = str(Path("destination") / "idle.png")

        # create the instance
        loader = BackgroundLoader(
            destination=Path("destination"),
            package="package",
            directory=Path("directory"),
        )

        # copy the backgrounds
        with self.assertLogs("dakara_player.background", "DEBUG") as logger:
            output_path = loader.copy_custom_background("idle", "idle.png")

        # assert the background
        self.assertEqual(output_path, Path("destination") / "idle.png")

        # assert logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.background:Loading custom idle "
                "background file 'idle.png'",
            ],
        )

        # assert the call of the mocked method
        mocked_copy.assert_called_with(
            Path("directory") / "idle.png", Path("destination")
        )

    @patch("dakara_player.background.copy", autospec=True)
    @patch("dakara_player.background.as_file", autospec=True)
    @patch("dakara_player.background.files", autospec=True)
    def test_copy_default_background(self, mocked_files, mocked_as_file, mocked_copy):
        """Test to copy a default background."""
        mocked_as_file.return_value.__enter__.return_value = (
            Path("package") / "idle.png"
        )
        mocked_copy.return_value = str(Path("destination") / "idle.png")

        # create the instance
        loader = BackgroundLoader(
            destination=Path("destination"),
            package="package",
        )

        # copy the backgrounds
        with self.assertLogs("dakara_player.background", "DEBUG") as logger:
            output_path = loader.copy_default_background("idle", "idle.png")

        # assert the background
        self.assertEqual(output_path, Path("destination") / "idle.png")

        # assert logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.background:Loading default "
                "idle background file 'idle.png'",
            ],
        )

        # assert the call of the mocked method
        mocked_files.assert_called_with("package")
        mocked_files.return_value.joinpath.assert_called_with("idle.png")
        mocked_copy.assert_called_with(
            Path("package") / "idle.png", Path("destination")
        )



from importlib.resources import as_file, files
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from dakara_player.audio import get_audio_files, is_audio_file


class IsAudioFileTestCase(TestCase):
    def test_mp3(self):
        """Test to detect a MP3 file."""
        with as_file(files("tests.resources").joinpath("song2.mp3")) as file:
            self.assertTrue(is_audio_file(file))

    def test_ass(self):
        """Test to not detect an ASS file."""
        with as_file(files("tests.resources").joinpath("song2.ass")) as file:
            self.assertFalse(is_audio_file(file))

    def test_mkv(self):
        """Test to not detect a MKV file."""
        with as_file(files("tests.resources").joinpath("song2.mkv")) as file:
            self.assertFalse(is_audio_file(file))


@patch("dakara_player.audio.is_audio_file", autospec=True)
@patch.object(Path, "glob", autospec=True)
class GetAudioFileTestCase(TestCase):
    def test_no_audio(self, mocked_glob, mocked_is_audio_file):
        """Test no audio file found."""
        file_video = Path("aa") / "file.mp4"
        mocked_glob.return_value = [file_video]

        self.assertEqual(len(get_audio_files(file_video)), 0)
        mocked_glob.assert_called_with(Path("aa"), "file.*")

    def test_one_audio(self, mocked_glob, mocked_is_audio_file):
        """Test one audio file found."""
        file_video = Path("aa") / "file.mp4"
        file_audio = Path("aa") / "file.mp3"
        mocked_glob.return_value = [file_video, file_audio]
        mocked_is_audio_file.return_value = True

        files_audio = get_audio_files(file_video)
        self.assertEqual(len(files_audio), 1)
        self.assertListEqual(files_audio, [file_audio])
        mocked_glob.assert_called_with(Path("aa"), "file.*")

    def test_two_audio(self, mocked_glob, mocked_is_audio_file):
        """Test one audio file found."""
        file_video = Path("aa") / "file.mp4"
        file_audio1 = Path("aa") / "file.mp3"
        file_audio2 = Path("aa") / "file.ogg"
        mocked_glob.return_value = [file_video, file_audio1, file_audio2]
        mocked_is_audio_file.return_value = True

        files_audio = get_audio_files(file_video)
        self.assertEqual(len(files_audio), 2)
        self.assertListEqual(files_audio, [file_audio1, file_audio2])
        mocked_glob.assert_called_with(Path("aa"), "file.*")



"""Generate text screens."""

import json
import logging
from importlib.resources import as_file, files
from pathlib import Path

from dakara_base.exceptions import DakaraError
from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PackageLoader

ICON_MAP_FILE = "line-awesome.json"

LINK_TYPE_NAMES = {
    "OP": "Opening",
    "ED": "Ending",
    "IN": "Insert song",
    "IS": "Image song",
}

logger = logging.getLogger(__name__)


class TextGenerator:
    """Generator for text screens.

    It populates text contents that are used for idle or transition screens. It
    uses Jinja under the hood, and can use templates from a custom directory,
    or from a fallback package.

    Example of use:

    >>> from pathlib import Path
    >>> generator = TextGenerator(
    ...     package="package",
    ...     directory=Path("directory"),
    ...     filenames={
    ...         "idle": "idle.ass",
    ...         "transition": "transition.ass",
    ...     },
    ... )
    >>> generator.load()
    >>> idle_screen_content = generator.get_text(
    ...     "idle",
    ...     {
    ...         "notes": [
    ...             "line1",
    ...             "line2",
    ...         ],
    ...     },
    ... )

    Args:
        package (str): Package checked for text templates by default.
        directory (pathlib.Path): Custom directory checked for text templates.
        filenames (dict): Dictionary of text templates filenames. The key is the
            template name, the value the template file name.

    Attributes:
        package (str): Package checked for text templates by default.
        directory (pathlib.Path): Custom directory checked for text templates.
        filenames (dict): Dictionary of text templates filenames. The key is the
            template name, the value the template file name.
        environment (jinja2.Environment): Environment for Jinja2.
        icon_map (dict): Map of icons. Keys are icon name, values are icon character.
    """

    def __init__(self, package, directory=None, filenames=None):
        self.package = package
        self.directory = directory or Path()
        self.filenames = filenames or {}

        # Jinja2 elements
        self.environment = None

        # icon map
        self.icon_map = {}

    def load(self):
        """Load the different parts of the class.

        Here are the actions with side effect.
        """
        # load icon mapping
        self.load_icon_map()

        # load templates
        self.load_templates()

    def load_icon_map(self):
        """Load the icon map."""
        with as_file(files("dakara_player.resources").joinpath(ICON_MAP_FILE)) as file:
            self.icon_map = json.loads(file.read_text())

    def load_templates(self):
        """Set up Jinja environment."""
        logger.debug("Loading text templates")

        # create loaders
        loaders = [
            FileSystemLoader(self.directory),
            PackageLoader(*separate_package_last_directory(self.package)),
        ]

        # create Jinja2 environment
        self.environment = Environment(loader=ChoiceLoader(loaders))

        # add filter for converting font icon name to character
        self.environment.filters["icon"] = self.convert_icon

        # add filter for work link type complete name
        self.environment.filters["link_type_name"] = self.convert_link_type_name

        # add filter for duration
        self.environment.filters["duration"] = self.convert_duration

        # check loaded templates
        for name, file_name in self.filenames.items():
            self.check_template(name, file_name)

    def get_environment_loaders(self):
        """Return the different loaders used by Jinja.

        Returns:
            list: List of the different loaders.
        """
        return self.environment.loader.loaders

    def check_template(self, template_name, file_name):
        """Check if a template is accessible either custom or default.

        Args:
            template_name (str): Name of the text template.
            file_name (str): Name of the text template file.

        Raises:
            TemplateNotFoundError: If the template can neither be found on
                custom loader, nor default loader.
        """
        loader_custom, loader_default = self.get_environment_loaders()

        if file_name in loader_custom.list_templates():
            logger.debug(
                "Loading custom %s text template file '%s'", template_name, file_name
            )
            return

        if file_name in loader_default.list_templates():
            logger.debug(
                "Loading default %s text template file '%s'", template_name, file_name
            )
            return

        raise TemplateNotFoundError(
            f"No {template_name} text template file found for '{file_name}'"
        )

    def convert_icon(self, name):
        """Convert the name of an icon to its code.

        Args:
            name (str): Name of the icon.

        Returns:
            str: Corresponding character.
        """
        if name is None:
            return ""

        return chr(int(self.icon_map.get(name, "0020"), 16))

    @staticmethod
    def convert_link_type_name(link_type):
        """Convert the short name of a link type to its long name.

        Args:
            link_type (str): Short name of the link type.

        Returns:
            str: Long name of the link type.
        """
        return LINK_TYPE_NAMES[link_type]

    @staticmethod
    def convert_duration(duration):
        """Format a duration in seconds.

        Args:
            duration (int): Duration in seconds.

        Returns:
            str: Formatted duration, either:

            - `H:MM:SS` if the duration exceeds one hour;
            - `M:SS` otherwise.
        """
        if duration is None:
            return ""

        try:
            hours, seconds = divmod(duration, 3600)
            minutes, seconds = divmod(seconds, 60)

        except TypeError:
            return str(duration)

        if hours > 0:
            return "{:d}:{:02d}:{:02d}".format(hours, minutes, seconds)

        return "{:d}:{:02d}".format(minutes, seconds)

    def get_text(self, template_name, data):
        """Generate the text for the desired template.

        Args:
            template_name (str): Name of the text template.
            data (dict): Values to pass to the template.

        Returns:
            str: Generated text from the template with provided values.
        """
        return self.environment.get_template(self.filenames[template_name]).render(
            **data
        )


def separate_package_last_directory(package):
    """Separate last directory from package.

    Args:
        package (str): Package.

    Returns:
        tuple:
            str: Package without last directory.
            str: Last directory.
    """
    *package_list, package_directory = package.split(".")
    return ".".join(package_list), package_directory


class TemplateNotFoundError(DakaraError, FileNotFoundError):
    """Error raised when a template cannot be found."""



from contextlib import ExitStack
from pathlib import Path
from queue import Queue
from threading import Event
from unittest import TestCase
from unittest.mock import MagicMock, patch

from packaging.version import Version

from dakara_player.media_player.base import (
    InvalidStateError,
    MediaPlayerNotAvailableError,
    VersionNotFoundError,
)
from dakara_player.media_player.mpv import (
    MediaPlayerMpv,
    MediaPlayerMpvOld,
    MediaPlayerMpvPost0330,
    MediaPlayerMpvPost0340,
    MpvTooOldError,
)
from tests.utils import get_temp_dir


class MediaPlayerMpvTestCase(TestCase):
    """Test the static methods of the abstract MediaPlayerMpv class."""

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_get_version_opensuse_postrelease(self, mocked_mpv_class):
        """Test to get the version from an openSUSE mpv build."""
        # mock the version of mpv
        mocked_mpv_class.return_value.mpv_version = (
            "mpv 0.32.0+git.20200402T120653.5824ac7d36"
        )

        # call the method
        version = MediaPlayerMpvOld.get_version()

        # assert the result
        self.assertEqual(version.base_version, "0.32.0")
        self.assertTrue(version.is_postrelease)

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_get_version_old(self, mocked_mpv_class):
        """Test to get the version from a release build of mpv <0.37."""
        # mock the version of mpv
        mocked_mpv_class.return_value.mpv_version = "mpv 0.32.0"

        # call the method
        version = MediaPlayerMpvOld.get_version()

        # assert the result
        self.assertEqual(version.base_version, "0.32.0")
        self.assertFalse(version.is_postrelease)

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_get_version_meson(self, mocked_mpv_class):
        """Test to get the version from a release build of mpv >=0.37."""
        # mock the version of mpv
        mocked_mpv_class.return_value.mpv_version = "mpv v0.37.0"

        # call the method
        version = MediaPlayerMpv.get_version()

        # assert the result
        self.assertEqual(version.base_version, "0.37.0")
        self.assertFalse(version.is_postrelease)

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_get_version_meson_git(self, mocked_mpv_class):
        """Test to get the version from a vcs build of mpv."""
        # mock the version of mpv
        mocked_mpv_class.return_value.mpv_version = "mpv v0.37.0-364-g2cc3bc12db"

        # call the method
        version = MediaPlayerMpv.get_version()

        # assert the result
        self.assertEqual(version.base_version, "0.37.0")
        self.assertTrue(version.is_postrelease)

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_get_version_not_found(self, mocked_mpv_class):
        """Test to get the mpv version when it is not available."""
        # mock the version of mpv
        mocked_mpv_class.return_value.mpv_version = "none"

        # call the method
        with self.assertRaisesRegex(VersionNotFoundError, "Unable to get mpv version"):
            MediaPlayerMpvOld.get_version()

    def test_get_class_from_version(self):
        """Test to get media player for various versions of mpv."""
        versions = [
            ("0.28.0", MediaPlayerMpvOld),
            ("0.33.0", MediaPlayerMpvPost0330),
            ("0.34.0", MediaPlayerMpvPost0340),
        ]

        for version, media_player_class in versions:
            self.assertIs(
                MediaPlayerMpv.get_class_from_version(Version(version)),
                media_player_class,
            )

    def test_raise_old_version(self):
        """Test to check that MpvTooOldError is raised on old versions of mpv."""
        version = "0.27.0"

        with self.assertRaises(MpvTooOldError):
            MediaPlayerMpv.get_class_from_version(Version(version))

    @patch.object(MediaPlayerMpv, "get_version")
    @patch.object(MediaPlayerMpv, "get_class_from_version")
    def test_from_version_config(
        self, mocked_get_class_from_version, mocked_get_version
    ):
        """Test to get version from config."""
        MediaPlayerMpv.from_version(
            None, None, {"mpv": {"force_version": "0.28.0"}}, None
        )

        mocked_get_class_from_version.assert_called_with(Version("0.28.0"))
        mocked_get_version.assert_not_called()

    @patch.object(MediaPlayerMpv, "get_version")
    @patch.object(MediaPlayerMpv, "get_class_from_version")
    def test_from_version_player(
        self, mocked_get_class_from_version, mocked_get_version
    ):
        """Test to get version from player."""
        mocked_get_version.return_value = Version("0.27.0")

        MediaPlayerMpv.from_version(None, None, {"mpv": {}}, None)

        mocked_get_class_from_version.assert_called_with(Version("0.27.0"))
        mocked_get_version.assert_called_with()

    @patch.object(MediaPlayerMpv, "get_version")
    @patch.object(MediaPlayerMpv, "get_class_from_version")
    def test_instanciate(self, mocked_get_class_from_version, mocked_get_version):
        """Test to instanciate media player mpv class."""
        mocked_get_class_from_version.return_value = MagicMock()

        MediaPlayerMpv.from_version(None, None, {}, "tmp")
        mocked_get_class_from_version.return_value.assert_called_with(
            None, None, {}, "tmp"
        )

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_is_available_ok_direct(self, mocked_mpv_class):
        """Test to get availability directly."""
        self.assertTrue(MediaPlayerMpv.is_available())

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_is_available_ok_indirect(self, mocked_mpv_class):
        """Test to get availability indirectly."""
        mocked_mpv_class.side_effect = [FileNotFoundError(), MagicMock()]
        self.assertTrue(MediaPlayerMpv.is_available())

    @patch("dakara_player.media_player.mpv.mpv", None)
    def test_is_available_ng_no_module(self):
        """Test to get inavailability if mpv module cannot be loaded."""
        self.assertFalse(MediaPlayerMpv.is_available())

    @patch("dakara_player.media_player.mpv.mpv.MPV")
    def test_is_available_ng(self, mocked_mpv_class):
        """Test to get inavailability."""
        mocked_mpv_class.side_effect = FileNotFoundError()
        self.assertFalse(MediaPlayerMpv.is_available())


class MediaPlayerMpvModelTestCase(TestCase):
    """Test the mpv player class unitary."""

    mpv_player_class = None

    def setUp(self):
        # create playlist entry ID
        self.id = 42

        # create playlist entry file path
        self.song_file_path = Path("file")
        self.subtitle_file_path = Path("file_sub")

        # create plàylist entry
        self.playlist_entry = {
            "id": self.id,
            "song": {"title": "Song title", "file_path": self.song_file_path},
            "owner": "me",
        }

    def get_instance(
        self,
        config=None,
    ):
        """Get a heavily mocked instance of the desired subclass of MediaPlayerMpv.

        Args:
            config (dict): Configuration passed to the constructor.

        Returns:
            tuple: Contains the following elements:
                MediaPlayerMpv: Instance;
                tuple: Contains the mocked objects:
                    unittest.mock.MagicMock: MPV object.
                    unittest.mock.MagicMock: BackgroundLoader object.
                    unittest.mock.MagicMock: TextGenerator object.
                tuple: Contains the mocked classes:
                    unittest.mock.MagicMock: MPV class.
                    unittest.mock.MagicMock: BackgroundLoader class.
                    unittest.mock.MagicMock: TextGenerator class.
        """
        config = config or {"kara_folder": get_temp_dir()}

        with ExitStack() as stack:
            mocked_instance_class = stack.enter_context(
                patch("dakara_player.media_player.mpv.mpv.MPV")
            )

            mocked_background_loader_class = stack.enter_context(
                patch("dakara_player.media_player.base.BackgroundLoader")
            )

            mocked_text_generator_class = stack.enter_context(
                patch("dakara_player.media_player.base.TextGenerator")
            )

            return (
                self.mpv_player_class(Event(), Queue(), config, Path("temp")),
                (
                    mocked_instance_class.return_value,
                    mocked_background_loader_class.return_value,
                    mocked_text_generator_class.return_value,
                ),
                (
                    mocked_instance_class,
                    mocked_background_loader_class,
                    mocked_text_generator_class,
                ),
            )

    def set_playlist_entry(self, mpv_player, started=True):
        """Set a playlist entry and make the player play it.

        Args:
            mpv_player (MediaPlayerMpv): Instance of the mpv player.
            started (bool): If True, make the player play the song.
        """
        mpv_player.playlist_entry = self.playlist_entry

        # create mocked transition
        mpv_player.playlist_entry_data["transition"].path = (
            get_temp_dir() / "transition.png"
        )

        # create mocked song
        mpv_player.playlist_entry_data["song"].path = (
            mpv_player.kara_folder_path / self.song_file_path
        )
        mpv_player.playlist_entry_data["song"].path_subtitle = (
            mpv_player.kara_folder_path / self.subtitle_file_path
        )

        # set media has started
        if started:
            mpv_player.player.path = mpv_player.playlist_entry_data["song"].path
            mpv_player.player.sub_files = [
                mpv_player.playlist_entry_data["song"].path_subtitle
            ]
            mpv_player.player.playlist = [
                {
                    "id": 1,
                    "filename": mpv_player.player.path,
                    "current": True,
                    "playing": True,
                }
            ]
            mpv_player.player.pause = False


class MediaPlayerMpvOldTestCase(MediaPlayerMpvModelTestCase):
    """Test the old mpv player class unitary."""

    mpv_player_class = MediaPlayerMpvOld

    @patch.object(MediaPlayerMpvOld, "is_available")
    def test_init_unavailable(self, mocked_is_available):
        """Test when mpv is not available."""
        mocked_is_available.return_value = False

        with self.assertRaisesRegex(
            MediaPlayerNotAvailableError, "mpv is not available"
        ):
            MediaPlayerMpvOld(Event(), Queue(), {}, Path("temp"))

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_get_timing(self, mocked_is_playing_this):
        """Test to get timing."""
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mocked_player.time_pos = 42.42
        mocked_is_playing_this.side_effect = lambda what: what == "song"

        self.assertEqual(mpv_player.get_timing(), 42)

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_get_timing_idle(self, mocked_is_playing_this):
        """Test to get timing when idle."""
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mocked_player.time_pos = 42.42
        mocked_is_playing_this.side_effect = lambda what: what == "idle"

        self.assertEqual(mpv_player.get_timing(), 0)

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_get_timing_transition(self, mocked_is_playing_this):
        """Test to get timing when in transition."""
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mocked_player.time_pos = 42.42
        mocked_is_playing_this.side_effect = lambda what: what == "transition"

        self.assertEqual(mpv_player.get_timing(), 0)

    @patch.object(MediaPlayerMpvOld, "get_version")
    def test_load_player(self, mocked_get_version):
        """Test to load the instance."""
        # create mock
        mocked_get_version.return_value = "0.32.0"

        # create instance
        mpv_player, _, _ = self.get_instance()

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.load_player()

        # assert the calls
        mocked_get_version.assert_called_with()

        # assert the logs
        self.assertListEqual(
            logger.output, ["INFO:dakara_player.media_player.mpv:mpv 0.32.0"]
        )

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_restart_transition(self, mocked_is_playing_this):
        """Test to restart on transition screen."""
        mpv_player, (mocked_instance, _, _), _ = self.get_instance()
        player = mocked_instance.media_player_new.return_value

        # mock
        mocked_is_playing_this.side_effect = lambda what: what == "transition"
        player.time_pos = 42

        # call method
        mpv_player.restart()

        # assert call
        self.assertEqual(player.time_pos, 42)

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    @patch.object(MediaPlayerMpvOld, "clear_playlist_entry")
    def test_skip_idle(self, mocked_clear_playlist_entry, mocked_is_playing_this):
        """Test to skip on idle screen."""
        mpv_player, (mocked_instance, _, _), _ = self.get_instance()

        # mock
        mocked_is_playing_this.side_effect = lambda what: what == "idle"

        # call method
        mpv_player.skip(True)

        # assert call
        mpv_player.clear_playlist_entry.assert_not_called()

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_rewind_transition(self, mocked_is_playing_this):
        """Test to rewind on transition screen."""
        mpv_player, (mocked_instance, _, _), _ = self.get_instance()
        player = mocked_instance.media_player_new.return_value

        # mock
        mocked_is_playing_this.side_effect = lambda what: what == "transition"
        player.time_pos = 42

        # call method
        mpv_player.rewind()

        # assert call
        self.assertEqual(player.time_pos, 42)

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_fast_forward_transition(self, mocked_is_playing_this):
        """Test to advance on transition screen."""
        mpv_player, (mocked_instance, _, _), _ = self.get_instance()
        player = mocked_instance.media_player_new.return_value

        # mock
        mocked_is_playing_this.side_effect = lambda what: what == "transition"
        player.time_pos = 42

        # call method
        mpv_player.fast_forward()

        # assert call
        self.assertEqual(player.time_pos, 42)

    @patch.object(MediaPlayerMpvOld, "clear_playlist_entry")
    @patch.object(MediaPlayerMpvOld, "play")
    def test_handle_end_file_transition(self, mocked_play, mocked_clear_playlist_entry):
        """Test end file callback for after a transition."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("finished", MagicMock())
        self.set_playlist_entry(mpv_player)
        mocked_player.playlist[0]["filename"] = str(get_temp_dir() / "transition.png")

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_end_file({"event": "end-file"})

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:File end callback called",
                "DEBUG:dakara_player.media_player.mpv:Will play '{}'".format(
                    get_temp_dir() / self.song_file_path
                ),
            ],
        )

        # assert the call
        mocked_play.assert_called_with("song")
        mocked_clear_playlist_entry.assert_not_called()
        mpv_player.callbacks["finished"].assert_not_called()

    @patch.object(MediaPlayerMpvOld, "clear_playlist_entry")
    @patch.object(MediaPlayerMpvOld, "play")
    def test_handle_end_file_song(self, mocked_play, mocked_clear_playlist_entry):
        """Test end file callback for after a song."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("finished", MagicMock())
        self.set_playlist_entry(mpv_player)

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_end_file({"event": "end-file"})

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.media_player.mpv:File end callback called"],
        )

        # assert the call
        mocked_play.assert_not_called()
        mocked_clear_playlist_entry.assert_called_with()
        mpv_player.callbacks["finished"].assert_called_with(self.playlist_entry["id"])

    @patch.object(MediaPlayerMpvOld, "play")
    def test_handle_end_file_song_too_fast(self, mocked_play):
        """Test song end callback after a song with instantaneous response
        from server.
        """
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback(
            "finished", lambda _: self.set_playlist_entry(mpv_player, started=False)
        )
        self.set_playlist_entry(mpv_player)

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG"):
            mpv_player.handle_end_file({"event": "end-file"})

        # post asserts
        self.assertTrue(mpv_player.errors.empty())
        self.assertIsNotNone(mpv_player.playlist_entry_data["song"].path)

        # assert the call
        mocked_play.assert_not_called()

    @patch.object(MediaPlayerMpvOld, "clear_playlist_entry")
    @patch.object(MediaPlayerMpvOld, "play")
    def test_handle_end_file_other(self, mocked_play, mocked_clear_playlist_entry):
        """Test end file callback for unknown state."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("finished", MagicMock())
        self.set_playlist_entry(mpv_player)
        mocked_player.playlist[0]["filename"] = str(get_temp_dir() / "other")

        self.assertFalse(mpv_player.stop.is_set())

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG"):
            mpv_player.handle_end_file({"event": "end-file"})

        self.assertTrue(mpv_player.stop.is_set())
        exception_class, exception, _ = mpv_player.errors.get()
        self.assertIs(exception_class, InvalidStateError)
        self.assertIn("End file on an undeterminated state", str(exception))

        # assert the call
        mocked_play.assert_not_called()
        mocked_clear_playlist_entry.assert_not_called()
        mpv_player.callbacks["finished"].assert_not_called()

    @patch.object(MediaPlayerMpvOld, "skip")
    def test_handle_log_message(self, mocked_skip):
        """Test log message callback."""
        # create instance
        mpv_player, _, _ = self.get_instance()
        self.set_playlist_entry(mpv_player)

        # mock the call
        mpv_player.set_callback("error", MagicMock())

        # call the method
        with (
            self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger,
            self.assertLogs("mpv", "DEBUG") as logger_mpv,
        ):
            mpv_player.handle_log_messages("fatal", "mpv.component", "error message")

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Log message callback called",
                "ERROR:dakara_player.media_player.mpv:Unable to play '{}'".format(
                    get_temp_dir() / self.song_file_path
                ),
            ],
        )
        self.assertListEqual(
            logger_mpv.output, ["CRITICAL:mpv:mpv.component: error message"]
        )

        # assert the call
        mpv_player.callbacks["error"].assert_called_with(
            self.playlist_entry["id"], "Unable to play current song: error message"
        )
        mocked_skip.assert_called_with()

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_handle_start_file_transition(self, mocked_is_playing_this):
        """Test start file callback for a transition."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("started_transition", MagicMock())
        mpv_player.set_callback("started_song", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create mocks
        mocked_is_playing_this.side_effect = lambda what: what == "transition"

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_start_file({})

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Start file callback called",
                "INFO:dakara_player.media_player.mpv:Playing transition for "
                "'Song title'",
            ],
        )

        # assert the call
        mpv_player.callbacks["started_transition"].assert_called_with(
            self.playlist_entry["id"]
        )
        mpv_player.callbacks["started_song"].assert_not_called()

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_handle_start_file_song(self, mocked_is_playing_this):
        """Test start file callback for a song."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("started_transition", MagicMock())
        mpv_player.set_callback("started_song", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create mocks
        mocked_is_playing_this.side_effect = lambda what: what == "song"

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_start_file({})

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Start file callback called",
                "INFO:dakara_player.media_player.mpv:Now playing 'Song title' "
                "('{}')".format(get_temp_dir() / self.song_file_path),
            ],
        )

        # assert the call
        mpv_player.callbacks["started_transition"].assert_not_called()
        mpv_player.callbacks["started_song"].assert_called_with(
            self.playlist_entry["id"]
        )

        # assert the call
        mpv_player.callbacks["started_transition"].assert_not_called()
        mpv_player.callbacks["started_song"].assert_called_with(
            self.playlist_entry["id"]
        )

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_handle_start_file_idle(self, mocked_is_playing_this):
        """Test start file callback for a idle."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("started_transition", MagicMock())
        mpv_player.set_callback("started_song", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create mocks
        mocked_is_playing_this.side_effect = lambda what: what == "idle"

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_start_file({})

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Start file callback called",
                "DEBUG:dakara_player.media_player.mpv:Playing idle screen",
            ],
        )

        # assert the call
        mpv_player.callbacks["started_transition"].assert_not_called()
        mpv_player.callbacks["started_song"].assert_not_called()

    @patch.object(MediaPlayerMpvOld, "is_playing_this")
    def test_handle_start_file_unknown(self, mocked_is_playing_this):
        """Test start file callback for an unknown state."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("started_transition", MagicMock())
        mpv_player.set_callback("started_song", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create mocks
        mocked_is_playing_this.return_value = False

        self.assertFalse(mpv_player.stop.is_set())

        # call the method
        mpv_player.handle_start_file({})

        self.assertTrue(mpv_player.stop.is_set())
        exception_class, exception, _ = mpv_player.errors.get()
        self.assertIs(exception_class, InvalidStateError)
        self.assertIn("Start file on an undeterminated state", str(exception))

        # assert the call
        mpv_player.callbacks["started_transition"].assert_not_called()
        mpv_player.callbacks["started_song"].assert_not_called()

    @patch.object(MediaPlayerMpvOld, "get_timing")
    def test_handle_pause(self, mocked_get_timing):
        """Test pause callback."""
        # create instance
        mpv_player, _, _ = self.get_instance()
        mpv_player.set_callback("paused", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create the mocks
        mocked_get_timing.return_value = 42

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_pause({})

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Pause callback called",
                "DEBUG:dakara_player.media_player.mpv:Paused",
            ],
        )

        # assert the call
        mpv_player.callbacks["paused"].assert_called_with(self.playlist_entry["id"], 42)
        mocked_get_timing.assert_called_with()

    @patch.object(MediaPlayerMpvOld, "get_timing")
    def test_handle_resumed(self, mocked_get_timing):
        """Test unpause callback."""
        # create instance
        mpv_player, _, _ = self.get_instance()
        mpv_player.set_callback("resumed", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create the mocks
        mocked_get_timing.return_value = 42

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_unpause({})

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Unpause callback called",
                "DEBUG:dakara_player.media_player.mpv:Resumed play",
            ],
        )

        # assert the call
        mpv_player.callbacks["resumed"].assert_called_with(
            self.playlist_entry["id"], 42
        )
        mocked_get_timing.assert_called_with()

    def test_init(self):
        """Test to initialize mpv player with custom config."""
        mpv_player, (mocked_player, _, _), _ = self.get_instance(
            {"mpv": {"key1": "value1"}}
        )
        self.assertEqual(mocked_player.key1, "value1")

    def test_play_invalid(self):
        """Test to play invalid action."""
        mpv_player, _, _ = self.get_instance()

        with self.assertRaisesRegex(ValueError, "Unexpected action to play: none"):
            mpv_player.play("none")

        mpv_player.player.play.assert_not_called()

    def test_play_no_song_path_subtitle(self):
        """Test to play a file no detected subtitle."""
        mpv_player, _, _ = self.get_instance()
        mpv_player.playlist_entry_data["song"].path = "test_file"
        mpv_player.playlist_entry_data["song"].path_subtitle = None

        mpv_player.play("song")

        mpv_player.player.play.assert_called_with("test_file")
        self.assertNotEqual(mpv_player.player.sub_files, [None])


class MediaPlayerMpvPost0330TestCase(MediaPlayerMpvModelTestCase):
    """Test the post 0.33.0 mpv player class unitary."""

    mpv_player_class = MediaPlayerMpvPost0330

    @patch.object(MediaPlayerMpvPost0330, "clear_playlist_entry")
    @patch.object(MediaPlayerMpvPost0330, "play")
    def test_handle_end_file_transition(self, mocked_play, mocked_clear_playlist_entry):
        """Test end file callback for after a transition."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("finished", MagicMock())
        self.set_playlist_entry(mpv_player)
        mocked_player.playlist[0]["filename"] = str(get_temp_dir() / "transition.png")

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_end_file(
                {"event": "end-file", "reason": "eof", "playlist_entry_id": 1}
            )

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:File end callback called",
                "DEBUG:dakara_player.media_player.mpv:Will play '{}'".format(
                    get_temp_dir() / self.song_file_path
                ),
            ],
        )

        # assert the call
        mocked_play.assert_called_with("song")
        mocked_clear_playlist_entry.assert_not_called()
        mpv_player.callbacks["finished"].assert_not_called()

    @patch.object(MediaPlayerMpvPost0330, "clear_playlist_entry")
    @patch.object(MediaPlayerMpvPost0330, "play")
    def test_handle_end_file_song(self, mocked_play, mocked_clear_playlist_entry):
        """Test end file callback for after a song."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("finished", MagicMock())
        self.set_playlist_entry(mpv_player)

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_end_file(
                {"event": "end-file", "reason": "eof", "playlist_entry_id": 1}
            )

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.media_player.mpv:File end callback called"],
        )

        # assert the call
        mocked_play.assert_not_called()
        mocked_clear_playlist_entry.assert_called_with()
        mpv_player.callbacks["finished"].assert_called_with(self.playlist_entry["id"])

    @patch.object(MediaPlayerMpvPost0330, "play")
    def test_handle_end_file_song_too_fast(self, mocked_play):
        """Test song end callback after a song with instantaneous response
        from server.
        """
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback(
            "finished", lambda _: self.set_playlist_entry(mpv_player, started=False)
        )
        self.set_playlist_entry(mpv_player)

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG"):
            mpv_player.handle_end_file(
                {"event": "end-file", "reason": "eof", "playlist_entry_id": 1}
            )

        # post asserts
        self.assertTrue(mpv_player.errors.empty())
        self.assertIsNotNone(mpv_player.playlist_entry_data["song"].path)

        # assert the call
        mocked_play.assert_not_called()

    @patch.object(MediaPlayerMpvPost0330, "clear_playlist_entry")
    @patch.object(MediaPlayerMpvPost0330, "play")
    def test_handle_end_file_other(self, mocked_play, mocked_clear_playlist_entry):
        """Test end file callback for unknown state."""
        # create instance
        mpv_player, (mocked_player, _, _), _ = self.get_instance()
        mpv_player.set_callback("finished", MagicMock())
        self.set_playlist_entry(mpv_player)
        mocked_player.playlist[0]["filename"] = str(get_temp_dir() / "other")

        self.assertFalse(mpv_player.stop.is_set())

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG"):
            mpv_player.handle_end_file(
                {"event": "end-file", "reason": "eof", "playlist_entry_id": 1}
            )

        self.assertTrue(mpv_player.stop.is_set())
        exception_class, exception, _ = mpv_player.errors.get()
        self.assertIs(exception_class, InvalidStateError)
        self.assertIn("End file on an undeterminated state", str(exception))

        # assert the call
        mocked_play.assert_not_called()
        mocked_clear_playlist_entry.assert_not_called()
        mpv_player.callbacks["finished"].assert_not_called()


class MediaPlayerMpvPost0340TestCase(MediaPlayerMpvModelTestCase):
    """Test the post 0.34.0 mpv player class unitary."""

    mpv_player_class = MediaPlayerMpvPost0340

    @patch.object(MediaPlayerMpvPost0340, "get_version")
    def test_load_player(self, mocked_get_version):
        """Test to load the instance."""
        # create mock
        mocked_get_version.return_value = "0.34.0"

        # create instance
        mpv_player, _, _ = self.get_instance()

        # pre assert
        self.assertFalse(mpv_player.player_data.get("initializing", False))

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.load_player()

        # assert outcome
        self.assertFalse(mpv_player.player_data.get("initializing", False))

        # assert the calls
        mocked_get_version.assert_called_with()

        # assert the logs
        self.assertListEqual(
            logger.output, ["INFO:dakara_player.media_player.mpv:mpv 0.34.0"]
        )

    @patch.object(MediaPlayerMpvPost0340, "get_timing")
    def test_handle_pause(self, mocked_get_timing):
        """Test pause callback."""
        # create instance
        mpv_player, _, _ = self.get_instance()
        mpv_player.set_callback("paused", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create the mocks
        mocked_get_timing.return_value = 42

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_pause("pause", True)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Pause callback called",
                "DEBUG:dakara_player.media_player.mpv:Paused",
            ],
        )

        # assert the call
        mpv_player.callbacks["paused"].assert_called_with(self.playlist_entry["id"], 42)
        mocked_get_timing.assert_called_with()

    @patch.object(MediaPlayerMpvPost0340, "get_timing")
    def test_handle_resumed(self, mocked_get_timing):
        """Test unpause callback."""
        # create instance
        mpv_player, _, _ = self.get_instance()
        mpv_player.set_callback("resumed", MagicMock())
        self.set_playlist_entry(mpv_player)

        # create the mocks
        mocked_get_timing.return_value = 42

        # call the method
        with self.assertLogs("dakara_player.media_player.mpv", "DEBUG") as logger:
            mpv_player.handle_pause("pause", False)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.media_player.mpv:Pause callback called",
                "DEBUG:dakara_player.media_player.mpv:Resumed play",
            ],
        )

        # assert the call
        mpv_player.callbacks["resumed"].assert_called_with(
            self.playlist_entry["id"], 42
        )
        mocked_get_timing.assert_called_with()



from importlib.resources import as_file, files
from pathlib import Path
from queue import Empty
from shutil import copy
from tempfile import TemporaryDirectory
from time import sleep
from unittest import TestCase


class TestCasePoller(TestCase):
    """Test class that can poll the state of tested player."""

    DELAY = 0.1

    @classmethod
    def wait_is_playing(cls, player, what=None, wait_extra=DELAY):
        """Wait for the player to be playing or to play something specifically.

        Use a polling loop. Note that after the condition is fulfilled, the
        function waits a little bit more.

        Args:
            player (dakara_player.media_player.base.MediaPlayer): Player.
            what (str): Action to wait for. If not provided, wait for the
                player to be actually playing.
            wait_extra (float): Time to wait at the end of the function.
        """
        if what:
            cls.wait(
                player, lambda: player.is_playing() and player.is_playing_this(what)
            )

        else:
            cls.wait(player, lambda: player.is_playing())

        sleep(wait_extra)

    @classmethod
    def wait_is_paused(cls, player, wait_extra=DELAY):
        """Wait for the player to be paused.

        Use a polling loop. Note that after the condition is fulfilled, the
        function waits a little bit more.

        Args:
            player (dakara_player.media_player.base.MediaPlayer): Player.
            wait_extra (float): Time to wait at the end of the function.
        """
        cls.wait(player, lambda: player.is_paused())

        sleep(wait_extra)

    @classmethod
    def wait(cls, player, condition_method):
        """Wait for a condition to be true safely.

        Args:
            condition_method (function): Function to call in loop in a try/except
                structure, so as to not break the loop in cause of error.
        """
        while True:
            # handle any player error
            if player.stop.is_set():
                try:
                    _, error, traceback = player.errors.get(5)
                    error.with_traceback(traceback)
                    raise error

                except Empty as error_empty:
                    raise RuntimeError("Unexpected error happened") from error_empty

            try:
                if condition_method():
                    return

            except OSError:
                pass

            sleep(cls.DELAY)


class TestCaseKara(TestCase):
    """Test class that creates a working kara folder."""

    def setUp(self):
        # create kara folder
        self.kara_folder = TemporaryDirectory()
        # resolve to prevent DOS short paths on Windows CI
        self.kara_folder_path = Path(self.kara_folder.name).resolve()

        # create subtitle
        with as_file(files("tests.resources").joinpath("song1.ass")) as file:
            self.subtitle1_path = Path(copy(file, self.kara_folder_path))

        with as_file(files("tests.resources").joinpath("song2.ass")) as file:
            self.subtitle2_path = Path(copy(file, self.kara_folder_path))

        # create song
        with as_file(files("tests.resources").joinpath("song1.mkv")) as file:
            self.song1_path = Path(copy(file, self.kara_folder_path))

        with as_file(files("tests.resources").joinpath("song2.mkv")) as file:
            self.song2_path = Path(copy(file, self.kara_folder_path))

        with as_file(files("tests.resources").joinpath("song3.avi")) as file:
            self.song3_path = Path(copy(file, self.kara_folder_path))

        # create audio
        with as_file(files("tests.resources").joinpath("song2.mp3")) as file:
            self.audio2_path = Path(copy(file, self.kara_folder_path))

        # create playlist entry
        self.playlist_entry1 = {
            "id": 42,
            "song": {
                "title": "Song 1",
                "file_path": str(self.song1_path),
                "duration": 60,
            },
            "owner": "me",
            "use_instrumental": False,
        }

        self.playlist_entry2 = {
            "id": 43,
            "song": {
                "title": "Song 2",
                "file_path": str(self.song2_path),
                "duration": 60,
            },
            "owner": "me",
            "use_instrumental": False,
        }

        self.playlist_entry3 = {
            "id": 44,
            "song": {
                "title": "Song 3",
                "file_path": str(self.song3_path),
                "duration": 60,
            },
            "owner": "me",
            "use_instrumental": False,
        }

    def tearDown(self):
        self.kara_folder.cleanup()


class TestCasePollerKara(TestCasePoller, TestCaseKara):
    """Test class that polls player and create kara folder."""



import os
import sys
from unittest import TestCase, skipIf, skipUnless
from unittest.mock import patch

try:
    import tkinter

except ImportError:
    tkinter = None

from dakara_player.window import (
    DummyWindowManager,
    TkWindowManager,
    WindowManagerNotAvailableError,
    get_window_manager_class,
)


class DummyWindowManagerTestCase(TestCase):
    """Test the dummy window manager."""

    @patch.object(DummyWindowManager, "is_available", return_value=False)
    def test_init_not_available(self, mocked_is_available):
        """Test to init an unavailable window manager."""
        with self.assertRaises(WindowManagerNotAvailableError):
            DummyWindowManager(None, None)

        mocked_is_available.assert_called_with()

    def test_open_close(self):
        """Test to open and close dummy window."""
        with DummyWindowManager():
            pass

    def test_get_id(self):
        """Test to get dummy window ID."""
        with DummyWindowManager() as window_manager:
            self.assertIsNone(window_manager.get_id())


@skipUnless(TkWindowManager.is_available(), "Tkinter not available")
@patch.object(TkWindowManager, "ACTUALIZE_INTERVAL", 100)
class TkWindowManagerTestCase(TestCase):
    """Test the Tkinter window manager."""

    @patch("dakara_player.window.tkinter", None)
    def test_is_available_no_tkinter(self):
        """Test to check availability of Tkinter window manager without Tkinter."""
        self.assertFalse(TkWindowManager.is_available())

    @skipIf(
        "linux" in sys.platform and "DISPLAY" not in os.environ,
        "No display detected on Linux",
    )
    def test_open_close(self):
        """Test to open and close Tk window."""
        with TkWindowManager(disabled=True):
            pass

    @skipIf(
        "linux" in sys.platform and "DISPLAY" not in os.environ,
        "No display detected on Linux",
    )
    def test_get_id(self):
        """Test to get Tk window ID."""
        with TkWindowManager(disabled=True) as window_manager:
            self.assertIsNotNone(window_manager.get_id())


class GetWindowManagerClassTestCase(TestCase):
    """Test the get_window_manager_class helper."""

    @patch.object(TkWindowManager, "is_available", return_value=True)
    def test_get_tk_window_manager(self, mocked_is_available):
        """Test to get the Tkinter window manager."""
        self.assertIs(get_window_manager_class(), TkWindowManager)

    @patch.object(TkWindowManager, "is_available", return_value=False)
    def test_get_dummy_window_manager(self, mocked_is_available):
        """Test to get the Tkinter window manager."""
        self.assertIs(get_window_manager_class(), DummyWindowManager)



"""HTTP and WebSocket clients for the Dakara server."""

import logging

from dakara_base.http_client import HTTPClient, authenticated
from dakara_base.utils import truncate_message
from dakara_base.websocket_client import WebSocketClient

logger = logging.getLogger(__name__)


class HTTPClientDakara(HTTPClient):
    """HTTP client connected to the Dakara server.

    Example of use:

    >>> http_client = HTTPClientDakara(
    ...     {
    ...         "address": "www.example.com",
    ...         "port": 8080,
    ...         "login": "player",
    ...         "password": "pass"
    ...     },
    ...     enpoint_prefix="api/",
    ...     mute_raise=True,
    ... )
    >>> http_client.authenticate()

    Args:
        config (dict): Config of the client.
        endpoint_prefix (str): Prefix of the endpoint, added to the URL.
        mute_raise (bool): If true, no exception will be raised when sending
            requests to the server (but authentication), only logged.
    """

    @authenticated
    def post_player_error(self, playlist_entry_id, message):
        """Report an error to the server.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.
            message (str): Error message.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug(
            "Telling the server that playlist entry %i cannot be played",
            playlist_entry_id,
        )

        self.post(
            endpoint="playlist/player/errors/",
            data={
                "playlist_entry_id": playlist_entry_id,
                "error_message": truncate_message(message, 255),
            },
            message_on_error="Unable to send player error to server",
        )

    @authenticated
    def put_status_finished(self, playlist_entry_id):
        """Report that a playlist entry has finished.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug(
            "Telling the server that playlist entry %i is finished", playlist_entry_id
        )

        self.put(
            endpoint="playlist/player/status/",
            data={"event": "finished", "playlist_entry_id": playlist_entry_id},
            message_on_error="Unable to report that a playlist entry has finished",
        )

    @authenticated
    def put_status_started_transition(self, playlist_entry_id):
        """Report that the transition of a playlist entry has started.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug(
            "Telling the server that the transition of playlist entry %i has started",
            playlist_entry_id,
        )

        self.put(
            endpoint="playlist/player/status/",
            data={
                "event": "started_transition",
                "playlist_entry_id": playlist_entry_id,
            },
            message_on_error=(
                "Unable to report that the transition of a playlist entry has started"
            ),
        )

    @authenticated
    def put_status_started_song(self, playlist_entry_id):
        """Report that the song of a playlist entry has started.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug(
            "Telling the server that the song of playlist entry %i has started",
            playlist_entry_id,
        )

        self.put(
            endpoint="playlist/player/status/",
            data={"event": "started_song", "playlist_entry_id": playlist_entry_id},
            message_on_error=(
                "Unable to report that the song of a playlist entry has started"
            ),
        )

    @authenticated
    def put_status_could_not_play(self, playlist_entry_id):
        """Report that a playlist entry could not play.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug(
            "Telling the server that the playlist entry %i could not play",
            playlist_entry_id,
        )

        self.put(
            endpoint="playlist/player/status/",
            data={"event": "could_not_play", "playlist_entry_id": playlist_entry_id},
            message_on_error="Unable to report that playlist entry could not play",
        )

    @authenticated
    def put_status_paused(self, playlist_entry_id, timing):
        """Report that the player is paused.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.
            timing (int): Progress of the player in seconds.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug("Telling the server that the player is paused")

        self.put(
            endpoint="playlist/player/status/",
            data={
                "event": "paused",
                "playlist_entry_id": playlist_entry_id,
                "timing": timing,
            },
            message_on_error="Unable to report that the player is paused",
        )

    @authenticated
    def put_status_resumed(self, playlist_entry_id, timing):
        """Report that the player resumed playing.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.
            timing (int): Progress of the player in seconds.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug("Telling the server that the player resumed playing")

        self.put(
            endpoint="playlist/player/status/",
            data={
                "event": "resumed",
                "playlist_entry_id": playlist_entry_id,
                "timing": timing,
            },
            message_on_error="Unable to report that the player resumed playing",
        )

    @authenticated
    def put_status_update_timing(self, playlist_entry_id, timing):
        """Report that the player updated its timing.

        Args:
            playlist_entry_id (int): ID of the playlist entry. Must not be
                `None`.
            timing (int): Progress of the player in seconds.

        Raises:
            AssertError: If `playlist_entry_id` is `None`.
        """
        assert playlist_entry_id is not None, "Entry with ID None is invalid"

        logger.debug("Telling the server that the player updated its timing")

        self.put(
            endpoint="playlist/player/status/",
            data={
                "event": "updated_timing",
                "playlist_entry_id": playlist_entry_id,
                "timing": timing,
            },
            message_on_error="Unable to report that the player updated its timing",
        )


class WebSocketClientDakara(WebSocketClient):
    """WebSocket client connected to the Dakara server.

    Example of use:

    >>> config = {
    ...     "address": "www.example.com",
    ...     "port": 8080,
    ...     "login": "player",
    ...     "password": "pass"
    ... }
    >>> http_client = HTTPClientDakara(
    ...     config,
    ...     enpoint_prefix="api/",
    ... )
    >>> http_client.authenticate()
    >>> header = http_client.get_token_header()
    >>> from thread import Event
    >>> from queue import Queue
    >>> stop = Event()
    >>> errors = Queue()
    >>> ws_client = WebSocketClientDakara(
    ...     stop,
    ...     errors,
    ...     config,
    ...     enpoint="ws/playlist",
    ...     header=header
    ... )
    >>> ws_client.run()

    Args:
        stop (Event): Event to stop the program.
        errors (Queue): Queue of errors.
        config (dict): Configuration for the server, the same as
            HTTPClientDakara.
        endpoint (str): Enpoint of the WebSocket connection, added to the URL.
        header (dict): Header containing the authentication token.
    """

    def set_default_callbacks(self):
        """Set all the default callbacks."""
        self.set_callback("idle", lambda: None)
        self.set_callback("playlist_entry", lambda playlist_entry: None)
        self.set_callback("command", lambda command: None)
        self.set_callback("connection_lost", lambda: None)

    def on_connected(self):
        """Callback when the connection is open."""
        self.send_ready()

    def on_connection_lost(self):
        """Callback when the connection is lost."""
        self.callbacks["connection_lost"]()

    def receive_idle(self, content):
        """Receive idle order.

        Args:
            content (dict): Dictionary of the event
        """
        logger.debug("Received idle order")
        self.callbacks["idle"]()

    def receive_playlist_entry(self, content):
        """Receive new playlist entry.

        Args:
            content (dict): Dictionary of the event
        """
        logger.debug("Received new playlist entry %i order", content["id"])
        self.callbacks["playlist_entry"](content)

    def receive_command(self, content):
        """Receive a command.

        Args:
            content (dict): Dictionary of the event
        """
        command = content["command"]
        logger.debug("Received command %s order", command)
        self.callbacks["command"](command)

    def send_ready(self):
        """Tell the server that the player is ready."""
        logger.debug("Telling the server that the player is ready")
        self.send("ready")



"""Manage audio files independently of the media player."""

import filetype


def get_audio_files(filepath):
    """Get audio files with the same name as provided file.

    Args:
        filepath (pathlib.Path): Path of the initial file.

    Returns:
        list of pathlib.Path: List of paths of audio files.
    """
    # list files with similar stem
    items = filepath.parent.glob(f"{filepath.stem}.*")
    return [item for item in items if item != filepath and is_audio_file(item)]


def is_audio_file(file_path):
    """Detect if a file is audio file based on standard magic numbers.

    Args:
        file_path (pathlib.Path): Path of the file to investigate.

    Returns:
        bool: `True` if the file is an audio file, `False` otherwise.
    """
    kind = filetype.guess(str(file_path))
    if not kind:
        return False

    maintype, _ = kind.mime.split("/")

    return maintype == "audio"



"""Manage MRL."""

from pathlib import Path
from urllib.parse import unquote, urlparse


def mrl_to_path(file_mrl):
    """Convert a MRL to a filesystem path.

    VLC stores file paths as MRL, we have to bring it back to a more classic
    looking path format.

    Args:
        file_mrl (str): Path to the resource within MRL format.

    Returns:
        pathlib.Path: Path to the resource.
    """
    # TODO Replace by pathlib.Path.from_uri() available in Python 3.13
    path_string = unquote(urlparse(file_mrl).path)

    # remove first '/' if a colon character is found like in '/C:/a/b'
    if path_string[0] == "/" and path_string[2] == ":":
        path_string = path_string[1:]

    return Path(path_string).resolve()


def path_to_mrl(file_path):
    """Convert a filesystem path to MRL.

    Args:
        file_path (pathlib.Path or str): Path to the resource.

    Returns:
        str: Path to the resource within MRL format.
    """
    return file_path.as_uri()



from copy import deepcopy
from queue import Queue
from threading import Event
from unittest import TestCase
from unittest.mock import ANY, patch

from dakara_player.media_player.mpv import MediaPlayerMpv
from dakara_player.media_player.vlc import MediaPlayerVlc
from dakara_player.player import DakaraPlayer, DakaraWorker, UnsupportedMediaPlayerError

CONFIG = {
    "player": {
        "kara_folder": "/some/path",
        "fullscreen": True,
        "vlc": {"media_parameters": None, "instance_parameters": None},
        "templates": None,
        "backgrounds": None,
        "durations": None,
    },
    "server": {
        "address": "www.example.com",
        "login": "player_login",
        "password": "player_password",
        "ssl": True,
        "reconnect_interval": 10,
    },
    "loglevel": "info",
}


class DakaraWorkerTestCase(TestCase):
    """Test the `DakaraWorker` class."""

    def test_init(self):
        """Test to create the object."""
        stop = Event()
        errors = Queue()

        # create Dakara worker
        with self.assertLogs("dakara_player.player", "DEBUG") as logger:
            DakaraWorker(stop, errors, CONFIG)

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.player:Starting Dakara worker"],
        )

    def test_get_media_player_class_vlc(self):
        """Test to get VLC as media player."""
        stop = Event()
        errors = Queue()
        config = deepcopy(CONFIG)
        config["player"]["player_name"] = "Vlc"

        worker = DakaraWorker(stop, errors, config)
        media_player_class = worker.get_media_player_class()

        self.assertIs(media_player_class, MediaPlayerVlc)

    def test_get_media_player_class_mpv(self):
        """Test to get MPV as media player."""
        stop = Event()
        errors = Queue()
        config = deepcopy(CONFIG)
        config["player"]["player_name"] = "Mpv"

        worker = DakaraWorker(stop, errors, config)
        media_player_class = worker.get_media_player_class()

        self.assertIs(media_player_class, MediaPlayerMpv.from_version)

    def test_get_media_player_class_default(self):
        """Test to get the default media player."""
        stop = Event()
        errors = Queue()

        worker = DakaraWorker(stop, errors, CONFIG)
        media_player_class = worker.get_media_player_class()

        self.assertIs(media_player_class, MediaPlayerVlc)

    def test_get_media_player_class_unsupported(self):
        """Test to get an unsupported media player."""
        stop = Event()
        errors = Queue()
        config = deepcopy(CONFIG)
        config["player"]["player_name"] = "Unknown"

        worker = DakaraWorker(stop, errors, config)

        with self.assertRaisesRegex(
            UnsupportedMediaPlayerError, "No media player for 'Unknown'"
        ):
            worker.get_media_player_class()

    @patch("dakara_player.player.TemporaryDirectory", autospec=True)
    @patch("dakara_player.player.FontLoader", autospec=True)
    @patch("dakara_player.player.MediaPlayerVlc", autospec=True)
    @patch("dakara_player.player.HTTPClientDakara", autospec=True)
    @patch(
        "dakara_player.player.WebSocketClientDakara",
        autospec=True,
    )
    @patch("dakara_player.player.DakaraManager", autospec=True)
    def test_run(
        self,
        mocked_dakara_manager_class,
        mocked_dakara_websocket_client_class,
        mocked_dakara_http_client_class,
        mocked_vlc_player_class,
        mocked_font_loader_class,
        mocked_temporary_directory_class,
    ):
        """Test a dummy run."""
        # create mock instances
        mocked_dakara_server_websocket = (
            mocked_dakara_websocket_client_class.return_value.__enter__.return_value
        )
        mocked_dakara_server_http = mocked_dakara_http_client_class.return_value
        mocked_dakara_server_http.get_token_header.return_value = "token"
        mocked_vlc_player = mocked_vlc_player_class.return_value.__enter__.return_value
        mocked_font_loader = (
            mocked_font_loader_class.return_value.__enter__.return_value
        )

        # create safe worker control objects
        stop = Event()
        errors = Queue()

        # create Dakara worker
        dakara_worker = DakaraWorker(stop, errors, CONFIG)

        # set the stop event
        stop.set()

        # call the method
        with patch.dict(
            "dakara_player.player.MEDIA_PLAYER_CLASSES",
            {"vlc": mocked_vlc_player_class},
        ):
            dakara_worker.run()

        # assert the call
        mocked_temporary_directory_class.assert_called_with(suffix=".dakara")
        mocked_font_loader_class.assert_called_with("dakara_player.resources.fonts")
        mocked_font_loader.load.assert_called_with()
        mocked_vlc_player_class.assert_called_with(stop, errors, CONFIG["player"], ANY)
        mocked_vlc_player.load.assert_called_with()
        mocked_dakara_http_client_class.assert_called_with(
            CONFIG["server"], endpoint_prefix="api/", mute_raise=True
        )
        mocked_dakara_server_http.authenticate.assert_called_with()
        mocked_dakara_server_http.get_token_header.assert_called_with()
        mocked_dakara_websocket_client_class.assert_called_with(
            stop,
            errors,
            CONFIG["server"],
            header="token",
            endpoint="ws/playlist/device/",
        )
        mocked_dakara_manager_class.assert_called_with(
            mocked_font_loader,
            mocked_vlc_player,
            mocked_dakara_server_http,
            mocked_dakara_server_websocket,
        )
        mocked_dakara_server_websocket.timer.start.assert_called_with()


class DakaraPlayerTestCase(TestCase):
    """Test the `DakaraPlayer` class."""

    def test_init(self):
        """Test to create the object."""
        with self.assertLogs("dakara_player.player", "DEBUG") as logger:
            DakaraPlayer(CONFIG)

        # assert effect on logs
        self.assertListEqual(logger.output, ["DEBUG:dakara_player.player:Started main"])

    @patch("dakara_player.player.check_version")
    def test_load(self, mocked_check_version):
        """Test to perform side-effect actions."""
        dakara_player = DakaraPlayer(CONFIG)
        dakara_player.load()

        # assert call
        mocked_check_version.assert_called_with()

    @patch.object(DakaraPlayer, "run_safe")
    def test_run(self, mocked_run_safe):
        """Test a dummy run."""
        dakara_player = DakaraPlayer(CONFIG)
        dakara_player.run()

        # assert the call
        mocked_run_safe.assert_called_with(DakaraWorker, CONFIG)



from unittest import TestCase
from unittest.mock import patch

from dakara_player.version import check_version


class CheckVersionTestCase(TestCase):
    """Test the version checker."""

    def test_check_version_release(self):
        """Test to display the version for a release."""
        with self.assertLogs("dakara_player.version", "DEBUG") as logger:
            with patch.multiple(
                "dakara_player.version", __version__="0.0.0", __date__="1970-01-01"
            ):
                check_version()

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            ["INFO:dakara_player.version:Dakara player 0.0.0 (1970-01-01)"],
        )

    def test_check_version_non_release(self):
        """Test to display the version for a non release."""
        with self.assertLogs("dakara_player.version", "DEBUG") as logger:
            with patch.multiple(
                "dakara_player.version",
                __version__="0.1.0-dev",
                __date__="1970-01-01",
            ):
                check_version()

        # assert effect on logs
        self.assertListEqual(
            logger.output,
            [
                "INFO:dakara_player.version:Dakara player 0.1.0-dev (1970-01-01)",
                "WARNING:dakara_player.version:"
                "You are running a dev version, use it at your own risks!",
            ],
        )
